# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from .dar_client import AddressType
from .dar_client import AsyncDARClient
from .dar_client import DARClient

__all__ = ["AsyncDARClient", "DARClient", "AddressType"]



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
import json
from typing import Any
from typing import AsyncGenerator
from typing import Dict
from typing import Generic
from typing import NoReturn
from typing import Optional
from typing import Type
from typing import TypeVar

import httpx
from gql.transport import AsyncTransport
from gql.transport import Transport
from gql.transport.exceptions import TransportAlreadyConnected
from gql.transport.exceptions import TransportClosed
from gql.transport.exceptions import TransportProtocolError
from gql.transport.exceptions import TransportServerError
from graphql import DocumentNode
from graphql import ExecutionResult
from graphql import print_ast
from structlog import get_logger

from ..graph.util import graphql_error_from_dict

logger = get_logger()

AnyHTTPXClient = TypeVar("AnyHTTPXClient", httpx.Client, httpx.AsyncClient)


class BaseHTTPXTransport(Generic[AnyHTTPXClient]):
    """
    This class is inspired heavily by GQL's AIOHTTPTransport and RequestsHTTPTransport,
    encapsulating the logic in smaller functions to allow for greater reuse.
    """

    def __init__(
        self,
        url: str,
        client_cls: Type[AnyHTTPXClient],
        client_args: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        HTTPX Transport for GQL.

        Args:
            url: The GraphQL server URL. Example: 'http://mo:5000/graphql'.
            client_cls: The HTTPX client class to instantiate.
            client_args: Dict of extra args passed to the HTTPX client.
        """
        self.url = url
        self.client_cls: Type[AnyHTTPXClient] = client_cls
        self.client_args = client_args
        self.client: Optional[AnyHTTPXClient] = None

    @property
    def session(self) -> Optional[AnyHTTPXClient]:
        """
        GQL is hardcoded to access 'session', but we prefer to work with 'client', as is
        custom with HTTPX.

        Returns: HTTPX client.
        """
        return self.client

    def _connect(self) -> None:
        """
        Create a HTTPX client as self.client. Should be cleaned with a call to close().
        """
        if self.client is not None:
            raise TransportAlreadyConnected("Transport is already connected")
        self.client = self.client_cls(**(self.client_args or {}))

    @staticmethod
    def _construct_payload(
        document: DocumentNode,
        variable_values: Optional[Dict[str, Any]] = None,
        operation_name: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Construct the GraphQL payload to be sent to the server.

        Args:
            document: GQL document.
            variable_values: Optional variable values to be used in the query.
            operation_name: Optional name of the query.

        Returns: The GraphQL payload.
        """
        query_str = print_ast(document)
        payload: Dict[str, Any] = {
            "query": query_str,
        }

        if variable_values:
            payload["variables"] = variable_values
        if operation_name:
            payload["operationName"] = operation_name

        logger.debug({"payload": json.dumps(payload)})

        return payload

    def _decode_response(
        self, response: httpx.Response, query: Optional[str] = None
    ) -> ExecutionResult:
        """
        Decodes raw GraphQL response from the server.

        Args:
            response: Raw HTTPX response.
            query: Optional request query. Used for better output in case of errors.

        Returns: graphql ExecutionResult, containing the result and potential errors.
        """
        logger.debug({"response": response.text})

        try:
            result = response.json()
        except Exception:  # noqa
            self._raise_response_error(response, "Not a JSON answer")

        if "errors" not in result and "data" not in result:
            self._raise_response_error(response, "No 'data' or 'errors' keys in answer")

        errors = None
        if "errors" in result:
            errors = [graphql_error_from_dict(e, query) for e in result["errors"]]

        return ExecutionResult(
            data=result.get("data"),
            errors=errors,
            extensions=result.get("extensions"),
        )

    @staticmethod
    def _raise_response_error(resp: httpx.Response, reason: str) -> NoReturn:
        """
        Raise the TransportError subclass corresponding to the exception encountered.
        Raises TransportServerError if the http status code is 400 or higher, and
        TransportProtocolError in all other cases.

        Args:
            resp: Raw HTTPX response.
            reason: Error message for protocol errors.

        Raises: Subclass of TransportError.
        """
        try:
            resp.raise_for_status()  # raises a HTTPError if response status is >400
        except httpx.HTTPStatusError as e:
            raise TransportServerError(str(e), e.response.status_code) from e

        raise TransportProtocolError(
            f"Server did not return a GraphQL result: {reason}: {resp.text}"
        )


class HTTPXTransport(BaseHTTPXTransport[httpx.Client], Transport):
    def __init__(
        self,
        url: str,
        client_cls: Type[httpx.Client] = httpx.Client,
        client_args: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(url, client_cls, client_args)

    def connect(self) -> None:
        """
        Create a HTTPX Client as self.client. Should be cleaned with a call to close().
        """
        return super()._connect()

    def close(self) -> None:
        """
        Close the client. GQL automatically calls this.
        """
        if self.client is not None:
            self.client.close()
        self.client = None

    def execute(
        self,
        document: DocumentNode,
        *args: Any,
        variable_values: Optional[Dict[str, Any]] = None,
        operation_name: Optional[str] = None,
        extra_args: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> ExecutionResult:
        """
        Execute the provided document AST against the configured remote server using the
        current client.

        Args:
            document: GQL document.
            variable_values: Optional variable values to be used in the query.
            operation_name: Optional name of the query.
            extra_args: Extra arguments, passed through to the request.

        Returns: graphql ExecutionResult, containing the result and potential errors.
        """
        if self.client is None:
            raise TransportClosed("Transport is not connected")

        payload = self._construct_payload(document, variable_values, operation_name)
        response = self.client.post(url=self.url, json=payload, **(extra_args or {}))
        return self._decode_response(response, query=payload["query"])


class AsyncHTTPXTransport(BaseHTTPXTransport[httpx.AsyncClient], AsyncTransport):
    def __init__(
        self,
        url: str,
        client_cls: Type[httpx.AsyncClient] = httpx.AsyncClient,
        client_args: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(url, client_cls, client_args)

    async def connect(self) -> None:
        """
        Create a HTTPX AsyncClient as self.client. Should be cleaned with a call to
        close(). Wraps sync function since GQL expects a coroutine for AsyncTransport.
        """
        return super()._connect()

    async def close(self) -> None:
        """
        Close the client. GQL automatically calls this.
        """
        if self.client is not None:
            await self.client.aclose()
        self.client = None

    async def execute(
        self,
        document: DocumentNode,
        variable_values: Optional[Dict[str, Any]] = None,
        operation_name: Optional[str] = None,
        extra_args: Optional[Dict[str, Any]] = None,
    ) -> ExecutionResult:
        """
        Execute the provided document AST against the configured remote server using the
        current client.

        Args:
            document: GQL document.
            variable_values: Optional variable values to be used in the query.
            operation_name: Optional name of the query.
            extra_args: Extra arguments, passed through to the request.

        Returns: graphql ExecutionResult, containing the result and potential errors.
        """
        if self.client is None:
            raise TransportClosed("Transport is not connected")

        payload = self._construct_payload(document, variable_values, operation_name)
        response = await self.client.post(
            url=self.url, json=payload, **(extra_args or {})
        )
        return self._decode_response(response, query=payload["query"])

    def subscribe(
        self,
        document: DocumentNode,
        variable_values: Optional[Dict[str, Any]] = None,
        operation_name: Optional[str] = None,
    ) -> AsyncGenerator[ExecutionResult, None]:  # pragma: no cover
        """
        Subscribe is not supported on HTTP.

        Raises: NotImplementedError.
        """
        raise NotImplementedError("The HTTP transport does not support subscriptions")



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional

from more_itertools import unzip


def dict_map(
    dicty: Dict,
    key_func: Optional[Callable[[Any], Any]] = None,
    value_func: Optional[Callable[[Any], Any]] = None,
) -> Dict:
    """Map a dictionary's keys and values.

    Similar to the built-in `map` function, but for dictionaries.

    Example:
        ```Python
        pow2 = lambda x: x**2
        input_dict = {1: 1, 2: 2, 3: 3}

        output_dict = dict_map(input_dict, value_func=pow2)
        assert output_dict == {1: 1, 2: 4, 3: 9})

        output_dict = dict_map(input_dict, key_func=pow2)
        assert output_dict == {1: 1, 4: 2, 9: 3})
        ```

    Args:
        dicty:
            The dictionary to transform keys and values from.
        key_func:
            The function to apply to each key in the dictionary.

            *Note: Care must be taken for mapped keys to not destructively
                   interfere with one another, i.e. the `key_func` should be a
                   bijective function.*
        value_func:
            The function to apply to each value in the dictionary.

    Raises:
        ValueError: Raised if provided the `key_func` was not a bijective function.

    Returns:
        A dict where the functions has been applied to keys and values respectively.
    """
    # Handle base-cases, i.e. empty dict and no transformation
    if not dicty:
        return dicty
    if key_func is None and value_func is None:
        return dicty

    keys, values = unzip(dicty.items())
    if key_func:
        keys = map(key_func, keys)
    if value_func:
        values = map(value_func, values)
    result_dict = dict(zip(keys, values))

    if len(result_dict) != len(dicty):
        raise ValueError("Provided `key_func` is non-bijective")
    return result_dict


def dict_map_key(key_func: Callable[[Any], Any], dicty: Dict) -> Dict:
    """Map a dictionary's keys.

    Similar to the built-in `map` function, but for dictionary keys.

    *Note: Utilizes `dict_map` internally, but swaps argument order to be more
           in line with the built-in `map` function.*

    Example:
        ```Python
        pow2 = lambda x: x**2
        input_dict = {1: 1, 2: 2, 3: 3}

        output_dict = dict_map_key(pow2, input_dict)
        assert output_dict == {1: 1, 4: 2, 9: 3}
        ```

    Args:
        dicty:
            The dictionary to transform keys and values from.
        key_func:
            The function to apply to each key in the dictionary.

            *Note: Care must be taken for mapped keys to not destructively
                   interfere with one another, i.e. the `key_func` should be a
                   bijective function.*

    Raises:
        ValueError: Raised if provided the `key_func` was not a bijective function.

    Returns:
        A dict where the `key_func` has been applied to every key.
    """
    return dict_map(dicty, key_func=key_func)


def dict_map_value(value_func: Callable[[Any], Any], dicty: Dict) -> Dict:
    """Map a dictionary's values.

    Similar to the built-in `map` function, but for dictionary values.

    *Note: Utilizes `dict_map` internally, but swaps argument order to be more
           in line with the built-in `map` function.*

    Example:
        ```Python
        pow2 = lambda x: x**2
        input_dict = {1: 1, 2: 2, 3: 3}

        output_dict = dict_map_value(pow2, input_dict)
        assert output_dict == {1: 1, 2: 4, 3: 9}
        ```

    Args:
        dicty:
            The dictionary to transform keys and values from.
        value_func:
            The function to apply to each value in the dictionary.

    Returns:
        A dict where the `value_func` has been applied to every value.
    """
    return dict_map(dicty, value_func=value_func)



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
"""FastRAMQPI Context."""

from typing import Any
from typing import AsyncContextManager
from typing import Awaitable
from typing import Callable
from typing import TypedDict

from authlib.integrations.httpx_client import AsyncOAuth2Client
from fastapi import FastAPI
from gql.client import AsyncClientSession
from prometheus_fastapi_instrumentator import Instrumentator
from pydantic import BaseSettings
from sqlalchemy.ext.asyncio import AsyncEngine
from sqlalchemy.ext.asyncio import async_sessionmaker

from .raclients.graph.client import GraphQLClient
from .raclients.modelclient.mo import ModelClient
from .ramqp.mo import MOAMQPSystem

HealthcheckFunction = Callable[["Context"], Awaitable[bool]]


class Context(TypedDict, total=False):
    """Execution context."""

    name: str
    settings: BaseSettings
    lifespan_managers: dict[int, set[AsyncContextManager]]
    app: FastAPI
    instrumentator: Instrumentator
    amqpsystem: MOAMQPSystem
    engine: AsyncEngine
    sessionmaker: async_sessionmaker

    # Legacy clients
    legacy_graphql_client: GraphQLClient
    legacy_graphql_session: AsyncClientSession
    legacy_model_client: ModelClient

    # Authenticated HTTPX Client
    mo_client: AsyncOAuth2Client

    # Autogenerated GraphQL Client
    # We cannot type the autogenerated GraphQL client since its signature depends on
    # the specific queries required by the integration.
    graphql_client: Any

    user_context: dict[str, Any]



from uuid import UUID

from .base_model import BaseModel


class DeclareEventListener(BaseModel):
    event_listener_declare: "DeclareEventListenerEventListenerDeclare"


class DeclareEventListenerEventListenerDeclare(BaseModel):
    uuid: UUID


DeclareEventListener.update_forward_refs()
DeclareEventListenerEventListenerDeclare.update_forward_refs()



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
# flake8: noqa
"""This module defines the public interface of the RAMQP package."""

from .amqp import AMQPSystem
from .amqp import PublishMixin
from .amqp import Router
from .utils import RejectMessage
from .utils import RequeueMessage



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
import traceback
from uuid import uuid4

import structlog
from fastapi.responses import PlainTextResponse
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from structlog.contextvars import bound_contextvars

logger = structlog.stdlib.get_logger()


class RequestIdMiddleware(BaseHTTPMiddleware):
    """Bind incoming X-Request-ID header to Structlog."""

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        request_id = request.headers.get("x-request-id", str(uuid4()))
        with bound_contextvars(request_id=request_id):
            return await call_next(request)


class ExceptionMiddleware(BaseHTTPMiddleware):
    """Properly log and return exceptions.

    Uncaught exceptions are normally logged by uvicorn, but that doesn't
    include our request id and other bound contextvars. Explicitly log
    exceptions and return the full traceback as the HTTP response.
    """

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        try:
            return await call_next(request)
        except Exception:
            logger.exception("Exception in application")
            return PlainTextResponse(content=traceback.format_exc(), status_code=500)



from typing import List
from uuid import UUID

from .base_model import BaseModel


class TestingCreateEmployee(BaseModel):
    employee_create: "TestingCreateEmployeeEmployeeCreate"


class TestingCreateEmployeeEmployeeCreate(BaseModel):
    validities: List["TestingCreateEmployeeEmployeeCreateValidities"]


class TestingCreateEmployeeEmployeeCreateValidities(BaseModel):
    uuid: UUID


TestingCreateEmployee.update_forward_refs()
TestingCreateEmployeeEmployeeCreate.update_forward_refs()
TestingCreateEmployeeEmployeeCreateValidities.update_forward_refs()



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from typing import Never


class TerminateTaskGroup(Exception):
    pass


async def terminate_task_group() -> Never:
    """Used to force termination of a task group.

    https://docs.python.org/3/library/asyncio-task.html#terminating-a-task-group
    """
    raise TerminateTaskGroup()



from typing import Any
from typing import Optional
from uuid import UUID

from ._testing__create_employee import TestingCreateEmployee
from ._testing__create_employee import TestingCreateEmployeeEmployeeCreate
from ._testing__get_employee import TestingGetEmployee
from ._testing__get_employee import TestingGetEmployeeEmployees
from ._testing__send_event import TestingSendEvent
from .acknowledge_event import AcknowledgeEvent
from .async_base_client import AsyncBaseClient
from .declare_event_listener import DeclareEventListener
from .declare_event_listener import DeclareEventListenerEventListenerDeclare
from .declare_event_namespace import DeclareEventNamespace
from .declare_event_namespace import DeclareEventNamespaceEventNamespaceDeclare
from .fetch_event import FetchEvent
from .fetch_event import FetchEventEventFetch
from .input_types import EventSendInput
from .input_types import ListenerCreateInput
from .input_types import NamespaceCreateInput


def gql(q: str) -> str:
    return q


class GraphQLClient(AsyncBaseClient):
    async def declare_event_namespace(
        self, input: NamespaceCreateInput
    ) -> DeclareEventNamespaceEventNamespaceDeclare:
        query = gql(
            """
            mutation DeclareEventNamespace($input: NamespaceCreateInput!) {
              event_namespace_declare(input: $input) {
                name
              }
            }
            """
        )
        variables: dict[str, object] = {"input": input}
        response = await self.execute(query=query, variables=variables)
        data = self.get_data(response)
        return DeclareEventNamespace.parse_obj(data).event_namespace_declare

    async def declare_event_listener(
        self, input: ListenerCreateInput
    ) -> DeclareEventListenerEventListenerDeclare:
        query = gql(
            """
            mutation DeclareEventListener($input: ListenerCreateInput!) {
              event_listener_declare(input: $input) {
                uuid
              }
            }
            """
        )
        variables: dict[str, object] = {"input": input}
        response = await self.execute(query=query, variables=variables)
        data = self.get_data(response)
        return DeclareEventListener.parse_obj(data).event_listener_declare

    async def fetch_event(self, listener: UUID) -> Optional[FetchEventEventFetch]:
        query = gql(
            """
            query FetchEvent($listener: UUID!) {
              event_fetch(filter: {listener: $listener}) {
                subject
                priority
                token
              }
            }
            """
        )
        variables: dict[str, object] = {"listener": listener}
        response = await self.execute(query=query, variables=variables)
        data = self.get_data(response)
        return FetchEvent.parse_obj(data).event_fetch

    async def acknowledge_event(self, token: Any) -> bool:
        query = gql(
            """
            mutation AcknowledgeEvent($token: EventToken!) {
              event_acknowledge(input: {token: $token})
            }
            """
        )
        variables: dict[str, object] = {"token": token}
        response = await self.execute(query=query, variables=variables)
        data = self.get_data(response)
        return AcknowledgeEvent.parse_obj(data).event_acknowledge

    async def _testing__send_event(self, input: EventSendInput) -> bool:
        query = gql(
            """
            mutation _Testing_SendEvent($input: EventSendInput!) {
              event_send(input: $input)
            }
            """
        )
        variables: dict[str, object] = {"input": input}
        response = await self.execute(query=query, variables=variables)
        data = self.get_data(response)
        return TestingSendEvent.parse_obj(data).event_send

    async def _testing__get_employee(
        self, cpr_number: Any
    ) -> TestingGetEmployeeEmployees:
        query = gql(
            """
            query _Testing_GetEmployee($cpr_number: CPR!) {
              employees(filter: {cpr_numbers: [$cpr_number]}) {
                objects {
                  validities {
                    uuid
                    cpr_number
                    given_name
                  }
                }
              }
            }
            """
        )
        variables: dict[str, object] = {"cpr_number": cpr_number}
        response = await self.execute(query=query, variables=variables)
        data = self.get_data(response)
        return TestingGetEmployee.parse_obj(data).employees

    async def _testing__create_employee(
        self, cpr_number: Any
    ) -> TestingCreateEmployeeEmployeeCreate:
        query = gql(
            """
            mutation _Testing_CreateEmployee($cpr_number: CPR!) {
              employee_create(
                input: {given_name: "Alice", surname: "Nielsen", cpr_number: $cpr_number}
              ) {
                validities {
                  uuid
                }
              }
            }
            """
        )
        variables: dict[str, object] = {"cpr_number": cpr_number}
        response = await self.execute(query=query, variables=variables)
        data = self.get_data(response)
        return TestingCreateEmployee.parse_obj(data).employee_create



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
import asyncio
from contextlib import asynccontextmanager
from dataclasses import dataclass
from dataclasses import field
from typing import AsyncIterator
from typing import Generic
from typing import TypeVar
from uuid import UUID
from uuid import uuid4

import structlog
from fastapi.encoders import jsonable_encoder
from httpx import AsyncClient
from httpx import ConnectError
from httpx import HTTPStatusError
from pydantic.generics import GenericModel
from structlog.contextvars import bound_contextvars

from .autogenerated_graphql_client import GraphQLClient
from .autogenerated_graphql_client import ListenerCreateInput
from .autogenerated_graphql_client import NamespaceCreateInput
from .config import Settings
from .util import TerminateTaskGroup
from .util import terminate_task_group

T = TypeVar("T")

logger = structlog.stdlib.get_logger()

# Define globally to allow overwriting sleep duration during testing
NO_EVENT_SLEEP_DURATION = 5.0


@dataclass(frozen=True)
class Namespace:
    """Namespace to declare in OS2mo.

    Only useful for integrations which generate their own events.
    """

    name: str
    public: bool = False


@dataclass(frozen=True)
class Listener:
    """Listener to declare in OS2mo.

    You can bind listeners to namespaces you did not declare yourself.
    """

    namespace: str
    user_key: str
    routing_key: str
    path: str
    parallelism: int = 1


@dataclass(frozen=True)
class GraphQLEvents:
    """FastRAMQPI configuration for the GraphQL event system."""

    declare_namespaces: list[Namespace] = field(default_factory=list)
    declare_listeners: list[Listener] = field(default_factory=list)


class Event(GenericModel, Generic[T], frozen=True):
    """A GraphQL event for use in FastAPI HTTP handlers."""

    subject: T
    priority: int


async def fetcher(
    integration_client: AsyncClient,
    graphql_client: GraphQLClient,
    listener: UUID,
    path: str,
    rate_limit_allowed: asyncio.Event,
    fetcher_number: int,
) -> None:
    log = logger.bind(listener=listener, n=fetcher_number)
    log.info("Starting fetcher")
    while True:
        request_id = uuid4()
        with bound_contextvars(request_id=request_id):
            try:
                # Wait until the event is set. This will pause the fetcher if
                # another fetcher for the same listener received a Retry-After.
                await rate_limit_allowed.wait()

                # Fetch GraphQL event from MO
                try:
                    event = await graphql_client.fetch_event(listener)
                except ConnectError:  # pragma: no cover
                    log.warning("Unable to fetch event (ConnectError)")
                    await asyncio.sleep(5)
                    continue
                log.debug("Fetched event", graphql_event=event)
                if event is None:
                    await asyncio.sleep(NO_EVENT_SLEEP_DURATION)
                    continue

                # HTTP POST event to the integration
                try:
                    r = await integration_client.post(
                        path,
                        headers={
                            "x-request-id": str(request_id),
                        },
                        # Pass all event arguments; we let the receiver decide which
                        # are important.
                        json=jsonable_encoder(
                            Event(
                                subject=event.subject,
                                priority=event.priority,
                            )
                        ),
                    )
                except ConnectError:  # pragma: no cover
                    log.warning("Unable to pass event to integration (ConnectError)")
                    await asyncio.sleep(5)
                    continue
                # 2xx is acknowledged, anything else is not. The GraphQL event
                # system does not have negative acknowledgements.
                try:
                    r.raise_for_status()
                except HTTPStatusError as e:
                    log.warning(
                        "HTTP status error in event callback",
                        status_code=e.response.status_code,
                        response=e.response.text,
                    )
                    # Rate-limiting
                    if retry_after := e.response.headers.get("Retry-After"):
                        # Retry-After can either be a HTTP-date (not ISO 8601!) or
                        # a number of seconds to delay. We only support the latter.
                        # https://datatracker.ietf.org/doc/html/rfc7231#section-7.1.3
                        # https://datatracker.ietf.org/doc/html/rfc7231#section-7.1.1.1
                        delay = int(retry_after)
                        log.warning("Rate-limited", delay=delay)
                        rate_limit_allowed.clear()  # pause all fetchers for this listener
                        await asyncio.sleep(delay)
                        rate_limit_allowed.set()  # resume all fetchers for this listener
                    continue
                log.debug("Acknowledging event", graphql_event=event)
                await graphql_client.acknowledge_event(event.token)
            except Exception:  # pragma: no cover
                log.exception("Unexpected exception in GraphQL event fetcher")
                await asyncio.sleep(5)


@asynccontextmanager
async def lifespan(
    settings: Settings,
    mo_client: AsyncClient,
    events: GraphQLEvents,
) -> AsyncIterator[None]:
    # The regular GraphQL client available in the FastRAMQPI context is
    # specific to each integration. We don't know which version of GraphQL it
    # it using, and we cannot define the required queries in it.
    # This client should probably be moved to the global context if FastRAMQPI
    # needs to execute other GraphQL queries in the future.
    graphql_client = GraphQLClient(
        url=f"{settings.mo_url}/graphql/v25",
        http_client=mo_client,
    )
    # HTTPX client to call the integration itself
    integration_client = AsyncClient(
        # We assume the integration is listening on port 8000
        base_url="http://127.0.0.1:8000",
        # Raise the timeout from the default of 5 seconds
        timeout=300,
    )
    logger.info("Starting GraphQL event fetchers")
    try:
        async with graphql_client, asyncio.TaskGroup() as tg:
            # Declare namespaces
            for namespace in events.declare_namespaces:
                logger.info("Declaring namespace", namespace=namespace)
                await graphql_client.declare_event_namespace(
                    input=NamespaceCreateInput(
                        name=namespace.name,
                        public=namespace.public,
                    )
                )
            # Declare listeners
            for listener in events.declare_listeners:
                logger.info("Declaring listener", listener=listener)
                graphql_listener = await graphql_client.declare_event_listener(
                    input=ListenerCreateInput(
                        namespace=listener.namespace,
                        user_key=listener.user_key,
                        routing_key=listener.routing_key,
                    )
                )
                # All fetchers for a listener share the same rate-limiting
                rate_limit_allowed = asyncio.Event()
                rate_limit_allowed.set()
                for i in range(listener.parallelism):
                    tg.create_task(
                        fetcher(
                            integration_client=integration_client,
                            graphql_client=graphql_client,
                            listener=graphql_listener.uuid,
                            path=listener.path,
                            rate_limit_allowed=rate_limit_allowed,
                            fetcher_number=i,
                        )
                    )
            yield
            logger.info("Stopping GraphQL event fetchers")
            tg.create_task(terminate_task_group())
    except* TerminateTaskGroup:
        pass



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
import hashlib
from functools import lru_cache
from functools import partial
from typing import Callable
from uuid import UUID


@lru_cache(maxsize=None)
def _generate_uuid(value: str) -> UUID:
    value_hash = hashlib.md5(value.encode())
    value_digest = value_hash.hexdigest()
    return UUID(value_digest)


def generate_uuid(base: str, value: str) -> UUID:
    """Generate a predictable uuid based on two variables.

    Example:
        ```Python
        uuid1 = uuid_generator("secret_seed", "facetnavn1")
        uuid2 = uuid_generator("secret_seed", "facetnavn1")
        uuid3 = uuid_generator("secret_seed", "facetnavn2")
        uuid4 = uuid_generator("s3cr3t_s33d", "facetnavn1")
        assert uuid1 == uuid2
        assert uuid1 != uuid3
        assert uuid1 != uuid4
        ```

    Args:
        base: Base or seed utilized to generate UUID.
        value: Specific evalue utilized to generate UUID.

    Returns:
        The generated UUID
    """
    base_uuid = _generate_uuid(base)
    return _generate_uuid(str(base_uuid) + str(value))


def uuid_generator(base: str) -> Callable[[str], UUID]:
    """Construct an UUID generator with a fixed base/seed.

    Example:
        ```Python
        uuid_gen = uuid_generator("secret_seed")
        uuid1 = uuid_gen("facetnavn1")
        uuid2 = uuid_gen("facetnavn1")
        uuid3 = uuid_gen("facetnavn2")
        assert uuid1 == uuid2
        assert uuid1 != uuid3
        ```

    Args:
        base: Base or seed utilized to generate all UUIDs.

    Returns:
        Function which map strings to UUIDs.
    """
    return partial(generate_uuid, base)



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from typing import Optional

from graphql import GraphQLError
from graphql import Source
from graphql import SourceLocation


def graphql_error_from_dict(d: dict, query: Optional[str] = None) -> GraphQLError:
    """
    Construct GraphQLError from error dict returned from the server.

    Args:
        d: Error dict.
        query: Original request query. Optional, but allows for better error messages.
    """
    error = GraphQLError(
        message=d["message"],
        # nodes=d.get("nodes"),
        source=Source(body=query) if query is not None else None,
        positions=d.get("positions"),
        path=d.get("path"),
        extensions=d.get("extensions"),
    )

    # For some reason, GraphQLError doesn't take locations in the constructor
    if error.locations is None and "locations" in d:
        error.locations = [SourceLocation(**loc) for loc in d["locations"]]

    return error



from typing import Any
from typing import List
from typing import Optional
from uuid import UUID

from .base_model import BaseModel


class TestingGetEmployee(BaseModel):
    employees: "TestingGetEmployeeEmployees"


class TestingGetEmployeeEmployees(BaseModel):
    objects: List["TestingGetEmployeeEmployeesObjects"]


class TestingGetEmployeeEmployeesObjects(BaseModel):
    validities: List["TestingGetEmployeeEmployeesObjectsValidities"]


class TestingGetEmployeeEmployeesObjectsValidities(BaseModel):
    uuid: UUID
    cpr_number: Optional[Any]
    given_name: str


TestingGetEmployee.update_forward_refs()
TestingGetEmployeeEmployees.update_forward_refs()
TestingGetEmployeeEmployeesObjects.update_forward_refs()
TestingGetEmployeeEmployeesObjectsValidities.update_forward_refs()



from .base_model import BaseModel


class DeclareEventNamespace(BaseModel):
    event_namespace_declare: "DeclareEventNamespaceEventNamespaceDeclare"


class DeclareEventNamespaceEventNamespaceDeclare(BaseModel):
    name: str


DeclareEventNamespace.update_forward_refs()
DeclareEventNamespaceEventNamespaceDeclare.update_forward_refs()



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
"""Pydantic StructuredUrl model."""

import json
from typing import Any
from typing import Optional
from urllib.parse import parse_qsl
from urllib.parse import quote
from urllib.parse import urlencode

from pydantic import AnyUrl
from pydantic import BaseModel
from pydantic import Field
from pydantic import SecretStr
from pydantic import parse_obj_as
from pydantic import root_validator


# pylint: disable=too-few-public-methods
class StructuredUrl(BaseModel):
    """Structured Url object.

    Allows for constructing a url either directly or indirectly."""

    class Config:
        """Settings are frozen."""

        frozen = True

    url: AnyUrl = Field(..., description="Database URL.")

    scheme: str | None
    user: str | None
    password: SecretStr | None
    host: str | None
    port: int | None
    path: str | None
    query: dict[str, str] | None
    fragment: str | None

    @root_validator(pre=True)
    def ensure_url(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Ensure that url is set.

        Args:
            values: Pydantic parsed values.

        Returns:
            'values' but with the guarantee that 'url' is set.
        """
        # If 'url' is set, noop.
        if values.get("url"):
            structured_field_keys = (
                "scheme",
                "user",
                "password",
                "host",
                "port",
                "path",
                "query",
                "fragment",
            )
            if not values.keys().isdisjoint(structured_field_keys):
                raise ValueError("cannot provide both url and structured fields")
            return values

        if "scheme" not in values:
            raise ValueError("scheme is required")
        if "host" not in values:
            raise ValueError("host is required")

        # Ensure that query is a dictionary or None
        query = values.get("query", {})
        if isinstance(query, str):
            query = json.loads(query)

        # According to RFC3986 section 3.2.1
        # Basic-auth username and password should be URL encoded
        user = values.get("user")
        if user:
            user = quote(user)
        password = values.get("password")
        if password:
            password = quote(password)

        uri_string = AnyUrl.build(
            scheme=values.get("scheme"),
            user=user,
            password=password,
            host=values.get("host"),
            port=parse_obj_as(Optional[str], values.get("port")),  # type: ignore[arg-type]
            path=values.get("path"),
            query=urlencode(query),
            fragment=values.get("fragment"),
        )
        values["url"] = parse_obj_as(AnyUrl, uri_string)
        return values

    @root_validator(pre=True)
    def ensure_xstructured_fields(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Ensure that our structured fields are set.

        Args:
            values: Pydantic parsed values.

        Returns:
            'values' but with the guarantee that all non-'url' fields are set.
        """
        # If 'url' is not set at this point, error out.
        assert values["url"] is not None

        url = parse_obj_as(AnyUrl, values["url"])
        values.update(
            dict(
                scheme=url.scheme,
                user=url.user,
                password=url.password,
                host=url.host,
                port=parse_obj_as(Optional[int], url.port),  # type: ignore[arg-type]
                path=url.path,
                query=dict(parse_qsl(url.query)),
                fragment=url.fragment,
            )
        )
        return values



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from functools import partial
from typing import Any
from typing import Callable

import tenacity
from tenacity import AsyncRetrying
from tenacity import WrappedFn
from tenacity import stop_after_delay
from tenacity import wait_fixed

retrying = partial(
    AsyncRetrying, stop=stop_after_delay(20), wait=wait_fixed(2), reraise=True
)


def retry(
    stop: Any = stop_after_delay(20),
    wait: Any = wait_fixed(2),
) -> Callable[[WrappedFn], WrappedFn]:
    """Tenacity retry decorator, with defaults useful for testing.

    Args:
        stop: Stop after 10 seconds so the test doesn't run forever.
            The default duration is long since some integrations need multiple rounds
            of AMQP messages to complete, and therefore need a long time to get in a
            consistent state.
        wait: Wait two seconds between assertion attempts.

    Returns:
        Function decorated with retrying.
    """
    # reraise=True so a false assertion fails the test
    return tenacity.retry(stop=stop, wait=wait, reraise=True)



from enum import Enum


class AccessLogModel(str, Enum):
    ACCESS_LOG = "ACCESS_LOG"
    PERSON = "PERSON"
    FACET = "FACET"
    IT_SYSTEM = "IT_SYSTEM"
    CLASS = "CLASS"
    ORGANISATION = "ORGANISATION"
    ORGANISATION_UNIT = "ORGANISATION_UNIT"
    ORGANISATION_FUNCTION = "ORGANISATION_FUNCTION"


class FileStore(str, Enum):
    EXPORTS = "EXPORTS"
    INSIGHTS = "INSIGHTS"


class OwnerInferencePriority(str, Enum):
    ENGAGEMENT = "ENGAGEMENT"
    ASSOCIATION = "ASSOCIATION"



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
import asyncio
from functools import wraps
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Optional
from typing import TypeVar

CallableReturnType = TypeVar("CallableReturnType")


def async_to_sync(
    func: Callable[..., Awaitable[CallableReturnType]],
) -> Callable[..., CallableReturnType]:
    """Function decorator to run an async function to completion.

    Example:
        ```Python
        @async_to_sync
        async def sleepy(seconds):
            await asyncio.sleep(seconds)
            return seconds

        print(sleepy(5))  # --> 5
        ```

    Args:
        func: The asynchronous function to wrap.

    Returns:
        The newly generated synchronous function wrapping the async one.
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Optional[Any]) -> CallableReturnType:
        return asyncio.run(func(*args, **kwargs))  # type: ignore

    return wrapper



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
import json
from functools import lru_cache
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import cast

# This path is now used on all (non-K8S) servers
_JSON_SETTINGS_PATH = "/opt/dipex/os2mo-data-import-and-export/settings/settings.json"


@lru_cache(maxsize=None)
def load_settings() -> Dict[str, Any]:
    """Load settings file from `settings/settings.json`.

    *Note: This function is in-memory cached using `lru_cache`, such that the
           underlying file is only read and parsed once, thus if the settings file is
           written to or updated after a program has called this function once, it
           will not return the new values.*

    If a refresh or reload is needed, the cache must first be invalidated:
    ```Python
    load_settings.clear_cache()
    ```

    Returns:
        The parsed `settings.json` file as a dictionary.
    """
    cwd = Path().cwd().absolute()
    settings_path = cwd / "settings" / "settings.json"

    if not Path(settings_path).exists():
        settings_path = Path(_JSON_SETTINGS_PATH)

    with open(str(settings_path), "r") as settings_file:
        return cast(Dict[str, Any], json.load(settings_file))


class Sentinel:
    def __str__(self) -> str:
        return "sentinel"

    def __repr__(self) -> str:
        return "sentinel"


read_setting_sentinel = Sentinel()


def load_setting(
    setting: str, default: Any = read_setting_sentinel
) -> Callable[[], Any]:
    """Load a single setting, defaulting to 'default' if not present

    This function is mainly for use as defaults in `click` or similar, hence why it is
    lazily evaluated.

    Example:
        With `settings.json` containing:
        ```Json
        {"a": 1}
        ```
        ```Python
        lazy_func = load_setting("a")
        assert lazy_func() == 1
        ```

    Example:
        ```Python
        import click
        from fastramqpi.ra_utils.load_settings import load_setting

        @click.command()
        @click.option(
            "--mora-base",
            default=load_setting("mora.base", "http://localhost:5000"),
            help="URL for OS2mo.",
        )
        def main(mora_base: str):
            ...
        ```

    Raises:
        ValueError: If the `setting` is not found and no `default` is set.

    Args:
        setting: The setting to read from `load_settings`.
        default: The default value to return, if `setting` is not found.

    Returns:
        A function evaluating to the value of the setting or the configured default.
    """

    def inner() -> Any:
        value = load_settings().get(setting, default)
        if value == read_setting_sentinel:
            raise ValueError(f"{setting} not in settings file and no default")
        return value

    return inner



from datetime import datetime
from typing import Any
from typing import Callable
from typing import Dict

from fastramqpi.ariadne import parse_graphql_datetime

SCALARS_PARSE_FUNCTIONS: Dict[Any, Callable[[Any], Any]] = {
    datetime: parse_graphql_datetime
}
SCALARS_SERIALIZE_FUNCTIONS: Dict[Any, Callable[[Any], Any]] = {}



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from collections.abc import Hashable
from decimal import Decimal
from typing import Any

from frozendict import frozendict

from .dict_map import dict_map


def is_hashable(value: Any) -> bool:
    """Check if input is hashable by attempting to hashing it.

    Args:
        value: The value to be checked for hashability.

    Returns:
        Whether the value is hashable or not.
    """
    try:
        hash(value)
    except TypeError:
        return False
    return True


def is_probably_hashable(value: Any) -> bool:
    """Check if input is probably hashable without hashing it.

    *Note: To get a definite answer use `is_hashable` instead.*

    Args:
        value: The value to be checked for probable hashability.

    Returns:
        Whether the value is probably hashable or not.
    """
    return isinstance(value, Hashable)


def ensure_hashable(value: Any) -> Any:
    """Convert input into hashable equivalents if required.

    Args:
        value: The value to make hashable.

    Raises:
        ValueError: If frozendict is not installed and a dict is provided.
        TypeError: If non-hashable and non-convertable types are provided.

    Returns:
        Hashable equivalent of value, i.e. frozenset if value is a set.
    """
    if isinstance(value, dict):
        value = frozendict(
            dict_map(
                value,
                key_func=ensure_hashable,
                value_func=ensure_hashable,
            )
        )
    elif isinstance(value, set):
        value = frozenset(map(ensure_hashable, value))
    elif isinstance(value, list):
        value = tuple(map(ensure_hashable, value))
    elif isinstance(value, Decimal) and value.is_snan():
        return Decimal("nan")
    elif isinstance(value, slice):
        # Builtin and not an acceptable base type, so cannot be extended
        # I.e. it is impossible to make a 'frozenslice'.
        raise TypeError("slice cannot be made hashable")
    if not is_hashable(value):
        raise TypeError(repr(value) + " is not hashable, please report this!")
    return value



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
import warnings
from functools import wraps
from typing import Any
from typing import Callable
from typing import Optional


def deprecated(func: Callable) -> Callable:
    """Function decorator to mark a function as deprecated.

    Example:
        ```Python
        @deprecated
        def old_function():
            pass

        old_function()
        # <stdin>:1: DeprecationWarning: Call to deprecated function func.
        ```

    Args:
        func: The function to mark as deprecated.

    Returns:
        The function which emits warnings when used.
    """

    @wraps(func)
    def new_func(*args: Any, **kwargs: Optional[Any]) -> Any:
        warnings.warn(
            "Call to deprecated function {}.".format(func.__name__),
            category=DeprecationWarning,
            stacklevel=2,
        )
        return func(*args, **kwargs)

    return new_func



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from asyncio import get_event_loop
from asyncio import iscoroutinefunction
from asyncio import new_event_loop
from functools import partial
from functools import wraps
from types import TracebackType
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

CoroutineReturnType = TypeVar("CoroutineReturnType")


class Syncable:
    """Helper mixin to support synchronized use of async classes.

    Must be given before the async class in inheritance hierarchy.

    *Note: Works by overridding `__getattribute__` to check for coroutine lookups
           and wrapping them in `asyncio.loop.run_until_complete` if an active
           event-loop is not detected.*

    Example:
        Basic usage:
        ```Python
        import asyncio

        class AsyncSleeper:
            async def sleep(self, time: int) -> int:
                await asyncio.sleep(time)
                return time

        class Sleeper(Syncable, AsyncSleeper):
            pass  # Syncable must come before AsyncSleeper

        async def async_call(time: int) -> int:
            return await Sleeper().sleep(time)

        def sync_call(time: int) -> int:
            return Sleeper().sleep(time)

        print(asyncio.run(async_call(1)))  # --> prints "1" after 1 second
        print(sync_call(2))                # --> prints "2" after 2 second
        ```

    Example:
        Context manager usage:
        ```Python
        import asyncio

        class AsyncContext:
            async def __aenter__(self):
                print("aenter")
                return self

            async def __aexit__(self, *err):
                print("aexit")

        class Context(Syncable, AsyncContext):
            pass  # Syncable must come before AsyncSleeper


        async def async_call() -> None:
            context = Context()
            async with context:
                print("inside")

        def sync_call() -> None:
            context = Context()
            with context:
                print("inside")

        # Both print "aenter", "inside", "aexit"
        asyncio.run(async_call())
        sync_call()
        ```
    """

    def __init__(self, *args: Any, **kwargs: Optional[Any]) -> None:
        super().__init__(*args, **kwargs)  # type: ignore
        try:
            self.__loop = get_event_loop()
        except RuntimeError:
            self.__loop = new_event_loop()

    def __enter__(self) -> None:
        """Call `__aenter__` if parent has it, `AttributeError` otherwise."""
        if hasattr(self, "__aenter__"):
            return self.__aenter__()  # type: ignore
        raise AttributeError("__enter__")

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> bool:
        """Call `__aexit__` if parent has it, `AttributeError` otherwise."""
        if hasattr(self, "__aexit__"):
            return self.__aexit__(exc_type, exc_value, exc_traceback)  # type: ignore
        raise AttributeError("__exit__")

    def __run_coroutine(
        self,
        coroutine: Callable[..., Awaitable[CoroutineReturnType]],
        *args: Any,
        **kwargs: Optional[Any],
    ) -> Union[Awaitable[CoroutineReturnType], CoroutineReturnType]:
        """Call coroutine if event-loop is running, call synchronized otherwise.

        Args:
            coroutine: The coroutine to execute either directly or synchronized.

        Returns:
            Awaitable if coroutine was executed directly, result otherwise.
        """
        if self.__loop.is_running():
            return coroutine(*args, **kwargs)
        return self.__loop.run_until_complete(coroutine(*args, **kwargs))

    def __getattribute__(self, name: str) -> Any:
        """Implementation of `__getattribute__` with special case for coroutines.

        Args:
            name: Name of the attribute to read.

        Returns:
            Exactly the same as `__getattribute__` normally does, unless the attribute
            is a coroutine, in which case a wrapped version is returned. Wrapped with
            `__run_coroutine`, to handle sync-calls outside of an active event-loop.
        """
        attribute = getattr(super(), name, None)
        if not attribute:
            return object.__getattribute__(self, name)
        if iscoroutinefunction(attribute):
            return wraps(attribute)(partial(self.__run_coroutine, attribute))
        return attribute



# This file has been modified by the UnsetInputTypesPlugin
from datetime import datetime
from typing import Any
from typing import List
from typing import Optional
from uuid import UUID

from pydantic import Field

from .base_model import UNSET
from .base_model import BaseModel
from .base_model import UnsetType
from .enums import AccessLogModel
from .enums import FileStore
from .enums import OwnerInferencePriority


class AccessLogFilter(BaseModel):
    ids: Optional[List[UUID]] = None
    uuids: Optional[List[UUID]] = None
    actors: Optional[List[UUID]] = None
    models: Optional[List[AccessLogModel]] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None


class AddressCreateInput(BaseModel):
    uuid: Optional[UUID] = None
    org_unit: Optional[UUID] = None
    person: Optional[UUID] = None
    employee: Optional[UUID] = None
    engagement: Optional[UUID] = None
    ituser: Optional[UUID] = None
    visibility: Optional[UUID] = None
    validity: "RAValidityInput"
    user_key: Optional[str] = None
    value: str
    address_type: UUID


class AddressFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    registration: Optional["AddressRegistrationFilter"] = None
    address_type: Optional["ClassFilter"] = None
    address_types: Optional[List[UUID]] = None
    address_type_user_keys: Optional[List[str]] = None
    engagement: Optional["EngagementFilter"] = None
    engagements: Optional[List[UUID]] = None
    ituser: Optional["ITUserFilter"] = None
    visibility: Optional["ClassFilter"] = None


class AddressRegistrationFilter(BaseModel):
    actors: Optional[List[UUID]] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None


class AddressTerminateInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: datetime
    uuid: UUID


class AddressUpdateInput(BaseModel):
    uuid: UUID
    org_unit: Optional[UUID] = None
    person: Optional[UUID] = None
    employee: Optional[UUID] = None
    engagement: Optional[UUID] = None
    ituser: Optional[UUID] = None
    visibility: Optional[UUID] = None
    validity: "RAValidityInput"
    user_key: Optional[str] = None
    value: Optional[str] = None
    address_type: Optional[UUID] = None


class AssociationCreateInput(BaseModel):
    uuid: Optional[UUID] = None
    user_key: Optional[str] = None
    primary: Optional[UUID] = None
    validity: "RAValidityInput"
    person: Optional[UUID] = None
    employee: Optional[UUID] = None
    substitute: Optional[UUID] = None
    trade_union: Optional[UUID] = None
    org_unit: UUID
    association_type: UUID


class AssociationFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    registration: Optional["AssociationRegistrationFilter"] = None
    association_type: Optional["ClassFilter"] = None
    association_types: Optional[List[UUID]] = None
    association_type_user_keys: Optional[List[str]] = None
    it_association: Optional[bool] = None


class AssociationRegistrationFilter(BaseModel):
    actors: Optional[List[UUID]] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None


class AssociationTerminateInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: datetime
    uuid: UUID


class AssociationUpdateInput(BaseModel):
    uuid: UUID
    user_key: Optional[str] = None
    primary: Optional[UUID] = None
    validity: "RAValidityInput"
    person: Optional[UUID] = None
    employee: Optional[UUID] = None
    substitute: Optional[UUID] = None
    trade_union: Optional[UUID] = None
    org_unit: Optional[UUID] = None
    association_type: Optional[UUID] = None


class ClassCreateInput(BaseModel):
    uuid: Optional[UUID] = None
    name: str
    user_key: str
    facet_uuid: UUID
    scope: Optional[str] = None
    published: str = "Publiceret"
    parent_uuid: Optional[UUID] = None
    example: Optional[str] = None
    owner: Optional[UUID] = None
    validity: "ValidityInput"
    it_system_uuid: Optional[UUID] = None
    description: Optional[str] = None


class ClassFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["ClassRegistrationFilter"] = None
    name: Optional[List[str]] = None
    facet: Optional["FacetFilter"] = None
    facets: Optional[List[UUID]] = None
    facet_user_keys: Optional[List[str]] = None
    parent: Optional["ClassFilter"] = None
    parents: Optional[List[UUID]] = None
    parent_user_keys: Optional[List[str]] = None
    it_system: Optional["ITSystemFilter"] = None
    owner: Optional["ClassOwnerFilter"] = None
    scope: Optional[List[str]] = None


class ClassOwnerFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["OrganisationUnitRegistrationFilter"] = None
    query: Optional[str] | UnsetType = UNSET
    names: Optional[List[str]] | UnsetType = UNSET
    parent: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    parents: Optional[List[UUID]] | UnsetType = UNSET
    child: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    hierarchy: Optional["ClassFilter"] = None
    hierarchies: Optional[List[UUID]] = None
    subtree: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    descendant: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    ancestor: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    engagement: Optional["EngagementFilter"] = None
    include_none: bool = False


class ClassRegistrationFilter(BaseModel):
    actors: Optional[List[UUID]] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None


class ClassTerminateInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: datetime
    uuid: UUID


class ClassUpdateInput(BaseModel):
    uuid: UUID
    name: str
    user_key: str
    facet_uuid: UUID
    scope: Optional[str] = None
    published: str = "Publiceret"
    parent_uuid: Optional[UUID] = None
    example: Optional[str] = None
    owner: Optional[UUID] = None
    validity: "ValidityInput"
    it_system_uuid: Optional[UUID] = None
    description: Optional[str] = None


class ConfigurationFilter(BaseModel):
    identifiers: Optional[List[str]] = None


class DescendantParentBoundOrganisationUnitFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["OrganisationUnitRegistrationFilter"] = None
    query: Optional[str] | UnsetType = UNSET
    names: Optional[List[str]] | UnsetType = UNSET
    parents: Optional[List[UUID]] | UnsetType = UNSET
    child: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    hierarchy: Optional["ClassFilter"] = None
    hierarchies: Optional[List[UUID]] = None
    subtree: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    ancestor: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    engagement: Optional["EngagementFilter"] = None


class EmployeeCreateInput(BaseModel):
    uuid: Optional[UUID] = None
    user_key: Optional[str] = None
    nickname_given_name: Optional[str] = None
    nickname_surname: Optional[str] = None
    seniority: Optional[Any] = None
    cpr_number: Optional[Any] = None
    given_name: str
    surname: str


class EmployeeFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["EmployeeRegistrationFilter"] = None
    query: Optional[str] | UnsetType = UNSET
    cpr_numbers: Optional[List[Any]] = None


class EmployeeRegistrationFilter(BaseModel):
    actors: Optional[List[UUID]] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None


class EmployeeTerminateInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: datetime
    uuid: UUID
    vacate: bool = False


class EmployeeUpdateInput(BaseModel):
    uuid: UUID
    user_key: Optional[str] = None
    nickname_given_name: Optional[str] = None
    nickname_surname: Optional[str] = None
    seniority: Optional[Any] = None
    cpr_number: Optional[Any] = None
    given_name: Optional[str] = None
    surname: Optional[str] = None
    validity: "RAValidityInput"


class EmployeesBoundAddressFilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["AddressRegistrationFilter"] = None
    address_type: Optional["ClassFilter"] = None
    address_types: Optional[List[UUID]] = None
    address_type_user_keys: Optional[List[str]] = None
    engagement: Optional["EngagementFilter"] = None
    engagements: Optional[List[UUID]] = None
    ituser: Optional["ITUserFilter"] = None
    visibility: Optional["ClassFilter"] = None


class EmployeesBoundAssociationFilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["AssociationRegistrationFilter"] = None
    association_type: Optional["ClassFilter"] = None
    association_types: Optional[List[UUID]] = None
    association_type_user_keys: Optional[List[str]] = None
    it_association: Optional[bool] = None


class EmployeesBoundEngagementFilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["EngagementRegistrationFilter"] = None
    job_function: Optional["ClassFilter"] = None
    engagement_type: Optional["ClassFilter"] = None


class EmployeesBoundITUserFilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["ITUserRegistrationFilter"] = None
    itsystem: Optional["ITSystemFilter"] = None
    itsystem_uuids: Optional[List[UUID]] = None
    engagement: Optional["EngagementFilter"] = None
    external_ids: Optional[List[str]] = None


class EmployeesBoundLeaveFilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["LeaveRegistrationFilter"] = None


class EmployeesBoundManagerFilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["ManagerRegistrationFilter"] = None
    responsibility: Optional["ClassFilter"] = None
    exclude: Optional["EmployeeFilter"] = None


class EngagementCreateInput(BaseModel):
    uuid: Optional[UUID] = None
    user_key: Optional[str] = None
    primary: Optional[UUID] = None
    validity: "RAValidityInput"
    extension_1: Optional[str] = None
    extension_2: Optional[str] = None
    extension_3: Optional[str] = None
    extension_4: Optional[str] = None
    extension_5: Optional[str] = None
    extension_6: Optional[str] = None
    extension_7: Optional[str] = None
    extension_8: Optional[str] = None
    extension_9: Optional[str] = None
    extension_10: Optional[str] = None
    employee: Optional[UUID] = None
    person: Optional[UUID] = None
    org_unit: UUID
    engagement_type: UUID
    job_function: UUID


class EngagementFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    registration: Optional["EngagementRegistrationFilter"] = None
    job_function: Optional["ClassFilter"] = None
    engagement_type: Optional["ClassFilter"] = None


class EngagementRegistrationFilter(BaseModel):
    actors: Optional[List[UUID]] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None


class EngagementTerminateInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: datetime
    uuid: UUID


class EngagementUpdateInput(BaseModel):
    uuid: UUID
    user_key: Optional[str] = None
    primary: Optional[UUID] = None
    validity: "RAValidityInput"
    extension_1: Optional[str] = None
    extension_2: Optional[str] = None
    extension_3: Optional[str] = None
    extension_4: Optional[str] = None
    extension_5: Optional[str] = None
    extension_6: Optional[str] = None
    extension_7: Optional[str] = None
    extension_8: Optional[str] = None
    extension_9: Optional[str] = None
    extension_10: Optional[str] = None
    employee: Optional[UUID] = None
    person: Optional[UUID] = None
    org_unit: Optional[UUID] = None
    engagement_type: Optional[UUID] = None
    job_function: Optional[UUID] = None


class EventAcknowledgeInput(BaseModel):
    token: Any


class EventFilter(BaseModel):
    listener: UUID


class EventSendInput(BaseModel):
    namespace: str
    routing_key: str
    subject: str
    priority: int = 10000


class EventSilenceInput(BaseModel):
    listeners: "ListenerFilter"
    subjects: List[str]


class EventUnsilenceInput(BaseModel):
    listeners: Optional["ListenerFilter"] = None
    subjects: Optional[List[str]] = None
    priorities: Optional[List[int]] = None


class FacetCreateInput(BaseModel):
    uuid: Optional[UUID] = None
    user_key: str
    published: str = "Publiceret"
    validity: "ValidityInput"


class FacetFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["FacetRegistrationFilter"] = None
    parent: Optional["FacetFilter"] = None
    parents: Optional[List[UUID]] = None
    parent_user_keys: Optional[List[str]] = None


class FacetRegistrationFilter(BaseModel):
    actors: Optional[List[UUID]] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None


class FacetTerminateInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: datetime
    uuid: UUID


class FacetUpdateInput(BaseModel):
    uuid: UUID
    user_key: str
    published: str = "Publiceret"
    validity: "ValidityInput"


class FacetsBoundClassFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["ClassRegistrationFilter"] = None
    name: Optional[List[str]] = None
    facet: Optional["FacetFilter"] = None
    facet_user_keys: Optional[List[str]] = None
    parent: Optional["ClassFilter"] = None
    parents: Optional[List[UUID]] = None
    parent_user_keys: Optional[List[str]] = None
    it_system: Optional["ITSystemFilter"] = None
    owner: Optional["ClassOwnerFilter"] = None
    scope: Optional[List[str]] = None


class FileFilter(BaseModel):
    file_store: FileStore
    file_names: Optional[List[str]] = None


class FullEventFilter(BaseModel):
    listeners: Optional["ListenerFilter"] = None
    subjects: Optional[List[str]] = None
    priorities: Optional[List[int]] = None
    silenced: Optional[bool] = None


class HealthFilter(BaseModel):
    identifiers: Optional[List[str]] = None


class ITAssociationCreateInput(BaseModel):
    uuid: Optional[UUID] = None
    user_key: Optional[str] = None
    primary: Optional[UUID] = None
    validity: "RAValidityInput"
    org_unit: UUID
    person: UUID
    it_user: UUID
    job_function: UUID


class ITAssociationTerminateInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: datetime
    uuid: UUID


class ITAssociationUpdateInput(BaseModel):
    uuid: UUID
    user_key: Optional[str] = None
    primary: Optional[UUID] = None
    validity: "RAValidityInput"
    org_unit: Optional[UUID] = None
    it_user: Optional[UUID] = None
    job_function: Optional[UUID] = None


class ITSystemCreateInput(BaseModel):
    uuid: Optional[UUID] = None
    user_key: str
    name: str
    validity: "RAOpenValidityInput"


class ITSystemFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["ITSystemRegistrationFilter"] = None


class ITSystemRegistrationFilter(BaseModel):
    actors: Optional[List[UUID]] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None


class ITSystemTerminateInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: datetime
    uuid: UUID


class ITSystemUpdateInput(BaseModel):
    uuid: UUID
    user_key: str
    name: str
    validity: "RAOpenValidityInput"


class ITUserCreateInput(BaseModel):
    uuid: Optional[UUID] = None
    external_id: Optional[str] = None
    primary: Optional[UUID] = None
    person: Optional[UUID] = None
    org_unit: Optional[UUID] = None
    engagement: Optional[UUID] = None
    validity: "RAValidityInput"
    user_key: str
    itsystem: UUID
    note: Optional[str] = None


class ITUserFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    registration: Optional["ITUserRegistrationFilter"] = None
    itsystem: Optional["ITSystemFilter"] = None
    itsystem_uuids: Optional[List[UUID]] = None
    engagement: Optional["EngagementFilter"] = None
    external_ids: Optional[List[str]] = None


class ITUserRegistrationFilter(BaseModel):
    actors: Optional[List[UUID]] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None


class ITUserTerminateInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: datetime
    uuid: UUID


class ITUserUpdateInput(BaseModel):
    uuid: UUID
    external_id: Optional[str] = None
    primary: Optional[UUID] = None
    person: Optional[UUID] = None
    org_unit: Optional[UUID] = None
    engagement: Optional[UUID] = None
    validity: "RAValidityInput"
    user_key: Optional[str] = None
    itsystem: Optional[UUID] = None
    note: Optional[str] = None


class ItuserBoundAddressFilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["AddressRegistrationFilter"] = None
    address_type: Optional["ClassFilter"] = None
    address_types: Optional[List[UUID]] = None
    address_type_user_keys: Optional[List[str]] = None
    engagement: Optional["EngagementFilter"] = None
    engagements: Optional[List[UUID]] = None
    visibility: Optional["ClassFilter"] = None


class ItuserBoundRoleBindingFilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["RoleRegistrationFilter"] = None


class KLECreateInput(BaseModel):
    uuid: Optional[UUID] = None
    user_key: Optional[str] = None
    org_unit: UUID
    kle_aspects: List[UUID]
    kle_number: UUID
    validity: "RAValidityInput"


class KLEFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    registration: Optional["KLERegistrationFilter"] = None


class KLERegistrationFilter(BaseModel):
    actors: Optional[List[UUID]] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None


class KLETerminateInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: datetime
    uuid: UUID


class KLEUpdateInput(BaseModel):
    uuid: UUID
    user_key: Optional[str] = None
    kle_number: Optional[UUID] = None
    kle_aspects: Optional[List[UUID]] = None
    org_unit: Optional[UUID] = None
    validity: "RAValidityInput"


class LeaveCreateInput(BaseModel):
    uuid: Optional[UUID] = None
    user_key: Optional[str] = None
    person: UUID
    engagement: UUID
    leave_type: UUID
    validity: "RAValidityInput"


class LeaveFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    registration: Optional["LeaveRegistrationFilter"] = None


class LeaveRegistrationFilter(BaseModel):
    actors: Optional[List[UUID]] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None


class LeaveTerminateInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: datetime
    uuid: UUID


class LeaveUpdateInput(BaseModel):
    uuid: UUID
    user_key: Optional[str] = None
    person: Optional[UUID] = None
    engagement: Optional[UUID] = None
    leave_type: Optional[UUID] = None
    validity: "RAValidityInput"


class ListenerCreateInput(BaseModel):
    namespace: str = "mo"
    user_key: str
    routing_key: str


class ListenerDeleteInput(BaseModel):
    uuid: UUID
    delete_pending_events: bool = False


class ListenerFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    owners: Optional[List[UUID]] = None
    routing_keys: Optional[List[str]] = None
    namespaces: Optional["NamespaceFilter"] = None


class ListenersBoundFullEventFilter(BaseModel):
    subjects: Optional[List[str]] = None
    priorities: Optional[List[int]] = None
    silenced: Optional[bool] = None


class ManagerCreateInput(BaseModel):
    uuid: Optional[UUID] = None
    user_key: Optional[str] = None
    person: Optional[UUID] = None
    responsibility: List[UUID]
    org_unit: UUID
    manager_level: UUID
    manager_type: UUID
    validity: "RAValidityInput"


class ManagerFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    registration: Optional["ManagerRegistrationFilter"] = None
    responsibility: Optional["ClassFilter"] = None
    exclude: Optional["EmployeeFilter"] = None


class ManagerRegistrationFilter(BaseModel):
    actors: Optional[List[UUID]] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None


class ManagerTerminateInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: datetime
    uuid: UUID


class ManagerUpdateInput(BaseModel):
    uuid: UUID
    validity: "RAValidityInput"
    user_key: Optional[str] = None
    person: Optional[UUID] = None
    responsibility: Optional[List[UUID]] = None
    org_unit: Optional[UUID] = None
    manager_type: Optional[UUID] = None
    manager_level: Optional[UUID] = None


class ModelsUuidsBoundRegistrationFilter(BaseModel):
    actors: Optional[List[UUID]] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None


class NamespaceCreateInput(BaseModel):
    name: str
    public: bool = False


class NamespaceDeleteInput(BaseModel):
    name: str


class NamespaceFilter(BaseModel):
    names: Optional[List[str]] = None
    owners: Optional[List[UUID]] = None
    public: Optional[bool] = None


class NamespacesBoundListenerFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    owners: Optional[List[UUID]] = None
    routing_keys: Optional[List[str]] = None


class OrgUnitsboundaddressfilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["AddressRegistrationFilter"] = None
    address_type: Optional["ClassFilter"] = None
    address_types: Optional[List[UUID]] = None
    address_type_user_keys: Optional[List[str]] = None
    engagement: Optional["EngagementFilter"] = None
    engagements: Optional[List[UUID]] = None
    ituser: Optional["ITUserFilter"] = None
    visibility: Optional["ClassFilter"] = None


class OrgUnitsboundassociationfilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["AssociationRegistrationFilter"] = None
    association_type: Optional["ClassFilter"] = None
    association_types: Optional[List[UUID]] = None
    association_type_user_keys: Optional[List[str]] = None
    it_association: Optional[bool] = None


class OrgUnitsboundengagementfilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["EngagementRegistrationFilter"] = None
    job_function: Optional["ClassFilter"] = None
    engagement_type: Optional["ClassFilter"] = None


class OrgUnitsboundituserfilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["ITUserRegistrationFilter"] = None
    itsystem: Optional["ITSystemFilter"] = None
    itsystem_uuids: Optional[List[UUID]] = None
    engagement: Optional["EngagementFilter"] = None
    external_ids: Optional[List[str]] = None


class OrgUnitsboundklefilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["KLERegistrationFilter"] = None


class OrgUnitsboundleavefilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["LeaveRegistrationFilter"] = None


class OrgUnitsboundmanagerfilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["ManagerRegistrationFilter"] = None
    responsibility: Optional["ClassFilter"] = None
    exclude: Optional["EmployeeFilter"] = None


class OrgUnitsboundrelatedunitfilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET


class OrganisationCreate(BaseModel):
    municipality_code: Optional[int] | UnsetType = UNSET


class OrganisationUnitCreateInput(BaseModel):
    uuid: Optional[UUID] | UnsetType = UNSET
    validity: "RAValidityInput"
    name: str
    user_key: Optional[str] | UnsetType = UNSET
    parent: Optional[UUID] | UnsetType = UNSET
    org_unit_type: UUID
    time_planning: Optional[UUID] | UnsetType = UNSET
    org_unit_level: Optional[UUID] | UnsetType = UNSET
    org_unit_hierarchy: Optional[UUID] | UnsetType = UNSET


class OrganisationUnitFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["OrganisationUnitRegistrationFilter"] = None
    query: Optional[str] | UnsetType = UNSET
    names: Optional[List[str]] | UnsetType = UNSET
    parent: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    parents: Optional[List[UUID]] | UnsetType = UNSET
    child: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    hierarchy: Optional["ClassFilter"] = None
    hierarchies: Optional[List[UUID]] = None
    subtree: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    descendant: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    ancestor: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    engagement: Optional["EngagementFilter"] = None


class OrganisationUnitRegistrationFilter(BaseModel):
    actors: Optional[List[UUID]] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None


class OrganisationUnitTerminateInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: datetime
    uuid: UUID


class OrganisationUnitUpdateInput(BaseModel):
    uuid: UUID
    validity: "RAValidityInput"
    name: Optional[str] | UnsetType = UNSET
    user_key: Optional[str] | UnsetType = UNSET
    parent: Optional[UUID] | UnsetType = UNSET
    org_unit_type: Optional[UUID] | UnsetType = UNSET
    org_unit_level: Optional[UUID] | UnsetType = UNSET
    org_unit_hierarchy: Optional[UUID] | UnsetType = UNSET
    time_planning: Optional[UUID] | UnsetType = UNSET


class OwnerCreateInput(BaseModel):
    uuid: Optional[UUID] = None
    user_key: Optional[str] = None
    org_unit: Optional[UUID] = None
    person: Optional[UUID] = None
    owner: Optional[UUID] = None
    inference_priority: Optional[OwnerInferencePriority] = None
    validity: "RAValidityInput"


class OwnerFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    owner: Optional["EmployeeFilter"] = None


class OwnerTerminateInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: datetime
    uuid: UUID


class OwnerUpdateInput(BaseModel):
    uuid: UUID
    user_key: Optional[str] = None
    org_unit: Optional[UUID] = None
    person: Optional[UUID] = None
    owner: Optional[UUID] = None
    inference_priority: Optional[OwnerInferencePriority] = None
    validity: "RAValidityInput"


class ParentsBoundClassFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["ClassRegistrationFilter"] = None
    name: Optional[List[str]] = None
    facet: Optional["FacetFilter"] = None
    facets: Optional[List[UUID]] = None
    facet_user_keys: Optional[List[str]] = None
    parent: Optional["ClassFilter"] = None
    parent_user_keys: Optional[List[str]] = None
    it_system: Optional["ITSystemFilter"] = None
    owner: Optional["ClassOwnerFilter"] = None
    scope: Optional[List[str]] = None


class ParentsBoundFacetFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["FacetRegistrationFilter"] = None
    parent: Optional["FacetFilter"] = None
    parent_user_keys: Optional[List[str]] = None


class ParentsBoundOrganisationUnitFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["OrganisationUnitRegistrationFilter"] = None
    query: Optional[str] | UnsetType = UNSET
    names: Optional[List[str]] | UnsetType = UNSET
    parent: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    child: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    hierarchy: Optional["ClassFilter"] = None
    hierarchies: Optional[List[UUID]] = None
    subtree: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    descendant: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    ancestor: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    engagement: Optional["EngagementFilter"] = None


class RAOpenValidityInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: Optional[datetime] = None


class RAValidityInput(BaseModel):
    from_: datetime = Field(alias="from")
    to: Optional[datetime] = None


class RegistrationFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    actors: Optional[List[UUID]] = None
    models: Optional[List[str]] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None


class RelatedUnitFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None


class RelatedUnitsUpdateInput(BaseModel):
    uuid: Optional[UUID] = None
    origin: UUID
    destination: Optional[List[UUID]] = None
    validity: "RAValidityInput"


class RoleBindingCreateInput(BaseModel):
    uuid: Optional[UUID] = None
    user_key: Optional[str] = None
    org_unit: Optional[UUID] = None
    ituser: UUID
    role: UUID
    validity: "RAValidityInput"


class RoleBindingFilter(BaseModel):
    uuids: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    registration: Optional["RoleRegistrationFilter"] = None
    ituser: Optional["ITUserFilter"] = None


class RoleBindingTerminateInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: datetime
    uuid: UUID


class RoleBindingUpdateInput(BaseModel):
    uuid: UUID
    user_key: Optional[str] = None
    org_unit: Optional[UUID] = None
    ituser: UUID
    role: Optional[UUID] = None
    validity: "RAValidityInput"


class RoleRegistrationFilter(BaseModel):
    actors: Optional[List[UUID]] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None


class UuidsBoundClassFilter(BaseModel):
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["ClassRegistrationFilter"] = None
    name: Optional[List[str]] = None
    facet: Optional["FacetFilter"] = None
    facets: Optional[List[UUID]] = None
    facet_user_keys: Optional[List[str]] = None
    parent: Optional["ClassFilter"] = None
    parents: Optional[List[UUID]] = None
    parent_user_keys: Optional[List[str]] = None
    it_system: Optional["ITSystemFilter"] = None
    owner: Optional["ClassOwnerFilter"] = None
    scope: Optional[List[str]] = None


class UuidsBoundEmployeeFilter(BaseModel):
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["EmployeeRegistrationFilter"] = None
    query: Optional[str] | UnsetType = UNSET
    cpr_numbers: Optional[List[Any]] = None


class UuidsBoundEngagementFilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["EngagementRegistrationFilter"] = None
    job_function: Optional["ClassFilter"] = None
    engagement_type: Optional["ClassFilter"] = None


class UuidsBoundFacetFilter(BaseModel):
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["FacetRegistrationFilter"] = None
    parent: Optional["FacetFilter"] = None
    parents: Optional[List[UUID]] = None
    parent_user_keys: Optional[List[str]] = None


class UuidsBoundITSystemFilter(BaseModel):
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["ITSystemRegistrationFilter"] = None


class UuidsBoundITUserFilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["ITUserRegistrationFilter"] = None
    itsystem: Optional["ITSystemFilter"] = None
    itsystem_uuids: Optional[List[UUID]] = None
    engagement: Optional["EngagementFilter"] = None
    external_ids: Optional[List[str]] = None


class UuidsBoundLeaveFilter(BaseModel):
    org_unit: Optional["OrganisationUnitFilter"] = None
    org_units: Optional[List[UUID]] = None
    employee: Optional["EmployeeFilter"] | UnsetType = UNSET
    employees: Optional[List[UUID]] = None
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["LeaveRegistrationFilter"] = None


class UuidsBoundOrganisationUnitFilter(BaseModel):
    user_keys: Optional[List[str]] = None
    from_date: Optional[datetime] | UnsetType = UNSET
    to_date: Optional[datetime] | UnsetType = UNSET
    registration: Optional["OrganisationUnitRegistrationFilter"] = None
    query: Optional[str] | UnsetType = UNSET
    names: Optional[List[str]] | UnsetType = UNSET
    parent: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    parents: Optional[List[UUID]] | UnsetType = UNSET
    child: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    hierarchy: Optional["ClassFilter"] = None
    hierarchies: Optional[List[UUID]] = None
    subtree: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    descendant: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    ancestor: Optional["OrganisationUnitFilter"] | UnsetType = UNSET
    engagement: Optional["EngagementFilter"] = None


class ValidityInput(BaseModel):
    from_: Optional[datetime] = Field(alias="from", default=None)
    to: Optional[datetime] = None


AccessLogFilter.update_forward_refs()
AddressCreateInput.update_forward_refs()
AddressFilter.update_forward_refs()
AddressRegistrationFilter.update_forward_refs()
AddressTerminateInput.update_forward_refs()
AddressUpdateInput.update_forward_refs()
AssociationCreateInput.update_forward_refs()
AssociationFilter.update_forward_refs()
AssociationRegistrationFilter.update_forward_refs()
AssociationTerminateInput.update_forward_refs()
AssociationUpdateInput.update_forward_refs()
ClassCreateInput.update_forward_refs()
ClassFilter.update_forward_refs()
ClassOwnerFilter.update_forward_refs()
ClassRegistrationFilter.update_forward_refs()
ClassTerminateInput.update_forward_refs()
ClassUpdateInput.update_forward_refs()
ConfigurationFilter.update_forward_refs()
DescendantParentBoundOrganisationUnitFilter.update_forward_refs()
EmployeeCreateInput.update_forward_refs()
EmployeeFilter.update_forward_refs()
EmployeeRegistrationFilter.update_forward_refs()
EmployeeTerminateInput.update_forward_refs()
EmployeeUpdateInput.update_forward_refs()
EmployeesBoundAddressFilter.update_forward_refs()
EmployeesBoundAssociationFilter.update_forward_refs()
EmployeesBoundEngagementFilter.update_forward_refs()
EmployeesBoundITUserFilter.update_forward_refs()
EmployeesBoundLeaveFilter.update_forward_refs()
EmployeesBoundManagerFilter.update_forward_refs()
EngagementCreateInput.update_forward_refs()
EngagementFilter.update_forward_refs()
EngagementRegistrationFilter.update_forward_refs()
EngagementTerminateInput.update_forward_refs()
EngagementUpdateInput.update_forward_refs()
EventAcknowledgeInput.update_forward_refs()
EventFilter.update_forward_refs()
EventSendInput.update_forward_refs()
EventSilenceInput.update_forward_refs()
EventUnsilenceInput.update_forward_refs()
FacetCreateInput.update_forward_refs()
FacetFilter.update_forward_refs()
FacetRegistrationFilter.update_forward_refs()
FacetTerminateInput.update_forward_refs()
FacetUpdateInput.update_forward_refs()
FacetsBoundClassFilter.update_forward_refs()
FileFilter.update_forward_refs()
FullEventFilter.update_forward_refs()
HealthFilter.update_forward_refs()
ITAssociationCreateInput.update_forward_refs()
ITAssociationTerminateInput.update_forward_refs()
ITAssociationUpdateInput.update_forward_refs()
ITSystemCreateInput.update_forward_refs()
ITSystemFilter.update_forward_refs()
ITSystemRegistrationFilter.update_forward_refs()
ITSystemTerminateInput.update_forward_refs()
ITSystemUpdateInput.update_forward_refs()
ITUserCreateInput.update_forward_refs()
ITUserFilter.update_forward_refs()
ITUserRegistrationFilter.update_forward_refs()
ITUserTerminateInput.update_forward_refs()
ITUserUpdateInput.update_forward_refs()
ItuserBoundAddressFilter.update_forward_refs()
ItuserBoundRoleBindingFilter.update_forward_refs()
KLECreateInput.update_forward_refs()
KLEFilter.update_forward_refs()
KLERegistrationFilter.update_forward_refs()
KLETerminateInput.update_forward_refs()
KLEUpdateInput.update_forward_refs()
LeaveCreateInput.update_forward_refs()
LeaveFilter.update_forward_refs()
LeaveRegistrationFilter.update_forward_refs()
LeaveTerminateInput.update_forward_refs()
LeaveUpdateInput.update_forward_refs()
ListenerCreateInput.update_forward_refs()
ListenerDeleteInput.update_forward_refs()
ListenerFilter.update_forward_refs()
ListenersBoundFullEventFilter.update_forward_refs()
ManagerCreateInput.update_forward_refs()
ManagerFilter.update_forward_refs()
ManagerRegistrationFilter.update_forward_refs()
ManagerTerminateInput.update_forward_refs()
ManagerUpdateInput.update_forward_refs()
ModelsUuidsBoundRegistrationFilter.update_forward_refs()
NamespaceCreateInput.update_forward_refs()
NamespaceDeleteInput.update_forward_refs()
NamespaceFilter.update_forward_refs()
NamespacesBoundListenerFilter.update_forward_refs()
OrgUnitsboundaddressfilter.update_forward_refs()
OrgUnitsboundassociationfilter.update_forward_refs()
OrgUnitsboundengagementfilter.update_forward_refs()
OrgUnitsboundituserfilter.update_forward_refs()
OrgUnitsboundklefilter.update_forward_refs()
OrgUnitsboundleavefilter.update_forward_refs()
OrgUnitsboundmanagerfilter.update_forward_refs()
OrgUnitsboundrelatedunitfilter.update_forward_refs()
OrganisationCreate.update_forward_refs()
OrganisationUnitCreateInput.update_forward_refs()
OrganisationUnitFilter.update_forward_refs()
OrganisationUnitRegistrationFilter.update_forward_refs()
OrganisationUnitTerminateInput.update_forward_refs()
OrganisationUnitUpdateInput.update_forward_refs()
OwnerCreateInput.update_forward_refs()
OwnerFilter.update_forward_refs()
OwnerTerminateInput.update_forward_refs()
OwnerUpdateInput.update_forward_refs()
ParentsBoundClassFilter.update_forward_refs()
ParentsBoundFacetFilter.update_forward_refs()
ParentsBoundOrganisationUnitFilter.update_forward_refs()
RAOpenValidityInput.update_forward_refs()
RAValidityInput.update_forward_refs()
RegistrationFilter.update_forward_refs()
RelatedUnitFilter.update_forward_refs()
RelatedUnitsUpdateInput.update_forward_refs()
RoleBindingCreateInput.update_forward_refs()
RoleBindingFilter.update_forward_refs()
RoleBindingTerminateInput.update_forward_refs()
RoleBindingUpdateInput.update_forward_refs()
RoleRegistrationFilter.update_forward_refs()
UuidsBoundClassFilter.update_forward_refs()
UuidsBoundEmployeeFilter.update_forward_refs()
UuidsBoundEngagementFilter.update_forward_refs()
UuidsBoundFacetFilter.update_forward_refs()
UuidsBoundITSystemFilter.update_forward_refs()
UuidsBoundITUserFilter.update_forward_refs()
UuidsBoundLeaveFilter.update_forward_refs()
UuidsBoundOrganisationUnitFilter.update_forward_refs()
ValidityInput.update_forward_refs()



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
import asyncio
import logging
import os
import urllib.parse
from asyncio import CancelledError
from asyncio import create_task
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from contextlib import suppress
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Iterator
from typing import Mapping
from typing import NoReturn
from unittest.mock import patch

# WARNING: You CANNOT import from `fastramqpi` here!
#
# This file is registered as a pytest plugin in pyproject.toml. Pytest does not
# allow controlling the order in which plugins are loaded, so if this plugin is
# loaded before pytest-cov, this file *and every line that was executed from
# its imports* is reported as missed because it doesn't execute after coverage
# collection is started.
# https://github.com/pytest-dev/pytest-cov/issues/635
# https://github.com/pytest-dev/pytest-cov/issues/682
# https://pytest-cov.readthedocs.io/en/latest/plugins.html
import httpx
import pytest
import sqlalchemy
import structlog
import uvicorn
from fastapi import FastAPI
from httpx import AsyncClient
from httpx import BasicAuth
from more_itertools import one
from pytest import Config
from pytest import FixtureRequest
from pytest import Item
from pytest import MonkeyPatch
from respx import MockRouter
from sqlalchemy import Connection
from sqlalchemy import text

# WARNING: Do not import from `fastramqpi` here!


logger = structlog.stdlib.get_logger()


def pytest_configure(config: Config) -> None:
    config.addinivalue_line(
        "markers", "integration_test: mark test as an integration test."
    )
    config.addinivalue_line(
        "markers", "envvar(mapping): set the specified environmental variables"
    )


def pytest_collection_modifyitems(items: list[Item]) -> None:
    """Automatically use convenient fixtures for tests marked with integration_test."""

    for item in items:
        if item.get_closest_marker("integration_test"):
            # MUST prepend to replicate auto-use fixtures coming first
            item.fixturenames[:0] = [  # type: ignore[attr-defined]
                "passthrough_backing_services",
                "fastramqpi_database_setup",
                "fastramqpi_database_isolation",
                "os2mo_database_snapshot_and_restore",
                "amqp_queue_isolation",
                "amqp_event_emitter",
                "graphql_events_quick_fetch",
                "graphql_events_quick_retry",
            ]
        else:  # unit-test
            # MUST prepend to replicate auto-use fixtures coming first
            item.fixturenames[:0] = [  # type: ignore[attr-defined]
                "empty_environment",
            ]


@pytest.fixture
async def empty_environment() -> AsyncIterator[None]:
    """Clear all environmental variables before running unit-test."""
    with patch.dict(os.environ, clear=True):
        yield


@pytest.fixture(autouse=True)
def load_marked_envvars(
    monkeypatch: pytest.MonkeyPatch,
    request: FixtureRequest,
) -> Iterator[None]:
    """Fixture to inject environmental variable via pytest.marks.

    Example:
        ```
        @pytest.mark.envvar({"VAR1": "1", "VAR2": 2})
        @pytest.mark.envvar({"VAR3": "3"})
        def test_load_marked_envvars() -> None:
            assert os.environ.get("VAR1") == "1"
            assert os.environ.get("VAR2") == "2"
            assert os.environ.get("VAR3") == "3"
            assert os.environ.get("VAR4") is None
        ```

    Args:
        monkeypatch: The patcher to use for settings the environmental variables.
        request: The pytest request object used to extract markers.

    Yields:
        None, but keeps the settings overrides active.
    """
    envvars: dict[str, str] = {}
    for mark in request.node.iter_markers("envvar"):
        if not mark.args:
            pytest.fail("envvar mark must take an argument")
        if len(mark.args) > 1:
            pytest.fail("envvar mark must take at most one argument")
        argument = one(mark.args)
        if not isinstance(argument, Mapping):
            pytest.fail("envvar mark argument must be a mapping")
        if any(not isinstance(key, str) for key in argument.keys()):
            pytest.fail("envvar mapping keys must be strings")
        if any(not isinstance(value, str) for value in argument.values()):
            pytest.fail("envvar mapping values must be strings")
        envvars.update(**argument)
    for key, value in envvars.items():
        monkeypatch.setenv(key, value)
    yield


@pytest.fixture(scope="session")
def _settings() -> Any:
    """Access FastRAMQPI settings without coupling to the integration's settings."""
    # We must defer importing from the FastRAMQPI module till run-time.
    # https://github.com/pytest-dev/pytest-cov/issues/587
    from fastramqpi.config import Settings

    class _Settings(Settings):
        class Config:
            env_prefix = "FASTRAMQPI__"

    return _Settings()


@pytest.fixture
async def unauthenticated_mo_client(_settings: Any) -> AsyncIterator[AsyncClient]:
    """HTTPX client with the OS2mo URL preconfigured."""
    mo_client = AsyncClient(base_url=_settings.mo_url)
    async with mo_client as client:
        yield client


@pytest.fixture
async def mo_client(_settings: Any) -> AsyncIterator[AsyncClient]:
    """HTTPX client with the OS2mo URL and auth preconfigured.

    For use by the integration's fixtures and tests. May be closed by the
    integration's ariadne codegen client, and as such cannot be used for
    post-test teardown (such as database restore).
    """
    from fastramqpi.main import construct_mo_client

    async with construct_mo_client(_settings) as client:
        yield client


@pytest.fixture
async def app() -> FastAPI:
    raise NotImplementedError(
        "You need to override the `app` fixture. See the FastRAMQPI README."
    )


def _is_uvicorn_cancellederror(record: logging.LogRecord) -> bool:
    if record.exc_info is None:
        return False
    exc_type, exc_value, exc_traceback = record.exc_info
    # https://github.com/Kludex/uvicorn/blob/918dae6ef91917a5824ab48c46026e07b7c80beb/uvicorn/server.py#L289
    if not isinstance(exc_value, CancelledError):
        return False
    if str(exc_value) != "Task cancelled, timeout graceful shutdown exceeded":
        return False
    logger.warning("Ignoring uvicorn cancellation errors")
    return True


@asynccontextmanager
async def run_server(app: FastAPI) -> AsyncIterator[None]:
    """Construct an ASGI server on localhost port 8000."""
    # Start uvicorn
    config = uvicorn.Config(
        app,
        host="127.0.0.1",
        port=8000,
        # When uvicorn is asked to shutdown, it first stops accepting new http
        # connections, and then waits for ongoing ones to finish, potentially
        # forever. Since the test has already passed (or failed!), we don't
        # want to spend any time waiting around.
        timeout_graceful_shutdown=0,
    )
    server = uvicorn.Server(config)
    task = None
    try:
        task = asyncio.create_task(server.serve())
        async with asyncio.timeout(60):
            while not server.started:
                await asyncio.sleep(0.1)
        # Yield the running server
        yield
    finally:
        # Uvicorn will throw a lot of errors when ongoing requests are
        # cancelled due to timeout_graceful_shutdown=0, ignore them.
        logging.getLogger("uvicorn.error").addFilter(
            lambda record: not _is_uvicorn_cancellederror(record)
        )

        # Always stop uvicorn whether our app exited successfully or not
        # The reason this is required here while being omitted for the other code in
        # this file, is that this is a standard asynccontextmanager while the functions
        # are pytest fixtures, for which pytest handles the exceptions
        server.should_exit = True

        # When uvicorn is asked to shutdown, it does the following:
        #  1. "Shutting down" (uvicorn stops accepting new http connections)
        #  2. "Waiting for connections to close"
        #  3. "Waiting for application shutdown" (uvicorn stops the lifespan)
        #  4. "Application shutdown complete"
        #  5. "Finished server process"
        # The duration of step 2 is controlled by the timeout_graceful_shutdown
        # option above, but step 3 might still take some time.
        assert task is not None
        await asyncio.wait_for(task, timeout=300)


@pytest.fixture
async def server(app: FastAPI) -> AsyncIterator[None]:
    async with run_server(app):
        yield


@asynccontextmanager
async def run_test_client() -> AsyncIterator[AsyncClient]:
    """Create httpx client bound to localhost port 8000."""
    # Yield HTTPX client for localhost
    async with AsyncClient(
        base_url="http://127.0.0.1:8000",
        # To avoid flaky tests, Starlette's TestClient does not enforce
        # timeouts[1]. We subscribe to their philosophy, but set a timeout of 5
        # minutes to avoid stuck tests.
        # [1] https://github.com/encode/starlette/issues/1108
        timeout=300,
    ) as client:
        yield client


@pytest.fixture
async def test_client(server: None) -> AsyncIterator[AsyncClient]:
    """Create httpx client bound to the local uvicorn."""
    async with run_test_client() as client:
        yield client


@pytest.fixture
async def rabbitmq_management_client(_settings: Any) -> AsyncIterator[AsyncClient]:
    """HTTPX client for the RabbitMQ management API."""
    amqp = _settings.amqp.get_url()
    async with httpx.AsyncClient(
        base_url=f"http://{amqp.host}:15672/api/",
        auth=BasicAuth(
            username=amqp.user,
            password=amqp.password,
        ),
    ) as client:
        yield client


@pytest.fixture(scope="session")
def superuser(_settings: Any) -> Iterator[Connection]:
    """Managing databases requires a superuser connection."""
    # Connect to "postgres" since we cannot drop a database while being connected to it.
    # TODO: it would be easier to use our own create_engine() from the database module,
    # but pytest cannot properly share event-loops across session-scoped fixtures and
    # (function-scoped) tests. Therefore, we use sqlalchemy's *sync* engine instead.
    # https://github.com/pytest-dev/pytest-asyncio/issues/706#issuecomment-1838860535
    db = _settings.database
    url = f"postgresql+psycopg://{db.user}:{db.password}@{db.host}:{db.port}/postgres"
    engine = sqlalchemy.create_engine(url)
    # AUTOCOMMIT disables transactions to allow for create/drop database operations
    engine.update_execution_options(isolation_level="AUTOCOMMIT")
    with engine.begin() as connection:
        yield connection


@pytest.fixture(scope="session")
def fastramqpi_database_setup(superuser: Connection) -> None:
    """Set up testing database template."""
    # Create separate testing template database. We will apply the database migrations
    # to this database once, and then use a copy of it for each test.
    template_db = "test_template"
    superuser.execute(text(f"drop database if exists {template_db}"))
    superuser.execute(text(f"create database {template_db}"))
    # Run migrations
    # TODO: alembic isn't implemented yet so we don't have to do anything here. Tables
    # are created on app startup for now.


@pytest.fixture
def fastramqpi_database_isolation(
    superuser: Connection, monkeypatch: MonkeyPatch
) -> None:
    """Ensure test isolation by resetting the database between tests.

    Automatically used on tests marked as integration_test.
    """
    # Copy template testing database (with migrations applied) to a temporary testing
    # database for the test that's about to run.
    template_db = "test_template"
    test_db = "test"
    superuser.execute(
        text(
            f"""
            select pg_terminate_backend(pid)
            from pg_stat_activity
            where datname = '{test_db}' and pid <> pg_backend_pid()
            """
        )
    )
    superuser.execute(text(f"drop database if exists {test_db}"))
    superuser.execute(text(f"create database {test_db} template {template_db}"))
    # Patch environment so the app under test will connect to this temporary database
    monkeypatch.setenv("FASTRAMQPI__DATABASE__NAME", test_db)


@pytest.fixture
async def amqp_emit_events(
    unauthenticated_mo_client: AsyncClient,
) -> Callable[[], Awaitable[None]]:
    """Emits outstanding events when called."""

    async def inner() -> None:
        r = await unauthenticated_mo_client.post("/testing/amqp/emit")
        r.raise_for_status()

    return inner


@pytest.fixture
async def amqp_event_emitter(
    amqp_emit_events: Callable[[], Awaitable[None]],
) -> AsyncIterator[None]:
    """Continuously, and quickly, emit OS2mo AMQP events during tests.

    Normally, OS2mo emits AMQP events periodically, but very slowly. Even though there
    are no guarantees as to message delivery speed, and we therefore should not design
    our system around such expectation, waiting a long time for tests to pass in the
    pipelines - or to fail during development - is a very poor development experience.

    Automatically used on tests marked as integration_test.
    """

    async def emitter() -> NoReturn:
        while True:
            await asyncio.sleep(1.13)
            await amqp_emit_events()

    task = create_task(emitter())
    yield
    task.cancel()
    with suppress(CancelledError):
        await task


@pytest.fixture
async def graphql_events_quick_retry(
    unauthenticated_mo_client: AsyncClient,
) -> AsyncIterator[None]:
    """Continuously reset GraphQL events to allow quick retrying during tests.

    Normally, events which are fetched - but not acknowledged - are not retried
    (cannot be fetched) for three minutes, increasing exponentially.

    Automatically used on tests marked as integration_test.
    """

    async def resetter() -> NoReturn:
        while True:
            await asyncio.sleep(0.941)
            r = await unauthenticated_mo_client.post("/testing/events/reset-last-tried")
            r.raise_for_status()

    task = create_task(resetter())
    yield
    task.cancel()
    with suppress(CancelledError):
        await task


@pytest.fixture
async def graphql_events_quick_fetch(monkeypatch: MonkeyPatch) -> None:
    """Quickly fetch GraphQL events during tests."""
    monkeypatch.setattr("fastramqpi.events.NO_EVENT_SLEEP_DURATION", 0.599)


@pytest.fixture
async def os2mo_database_snapshot_and_restore(
    unauthenticated_mo_client: AsyncClient,
) -> AsyncIterator[None]:
    """Ensure test isolation by resetting the OS2mo database between tests.

    Automatically used on tests marked as integration_test.
    """
    r = await unauthenticated_mo_client.post("/testing/database/snapshot")
    r.raise_for_status()
    yield
    r = await unauthenticated_mo_client.post("/testing/database/restore")
    r.raise_for_status()


@pytest.fixture
async def amqp_queue_isolation(
    rabbitmq_management_client: AsyncClient,
) -> None:
    """Ensure test isolation by deleting all AMQP queues before tests.

    Automatically used on tests marked as integration_test.
    """
    queues = (await rabbitmq_management_client.get("queues")).json()
    # vhost and name must be URL-encoded. This includes `/`, which is normally regarded
    # as safe. This is particularly important for the default AMQP vhost `/`.
    urls = (
        "queues/{vhost}/{name}".format(
            vhost=urllib.parse.quote(q["vhost"], safe=""),
            name=urllib.parse.quote(q["name"], safe=""),
        )
        for q in queues
    )
    deletes = [rabbitmq_management_client.delete(url) for url in urls]
    await asyncio.gather(*deletes)


@pytest.fixture
def passthrough_backing_services(_settings: Any, respx_mock: MockRouter) -> None:
    """Allow calls to the backing services to bypass the RESPX mocking.

    Automatically used on tests marked as integration_test.
    """
    # the integration itself, running in uvicorn
    respx_mock.route(host="127.0.0.1", port=8000).pass_through()
    # mo and keycloak are named to allow tests to revert the passthrough if needed
    respx_mock.route(name="keycloak", host=_settings.auth_server.host).pass_through()
    respx_mock.route(name="mo", host=_settings.mo_url.host).pass_through()
    # rabbitmq management
    respx_mock.route(host=_settings.amqp.get_url().host).pass_through()
    respx_mock.route(host="localhost").pass_through()


@pytest.fixture
def get_num_queued_messages(
    rabbitmq_management_client: AsyncClient,
) -> Callable[[], Awaitable[int]]:
    """Get number of queued messages in RabbitMQ AMQP."""

    async def _get_num_queued_messages() -> int:
        queues = (await rabbitmq_management_client.get("queues")).json()
        return sum(
            queue.get("messages_ready", 0) + queue.get("messages_unacknowledged", 0)
            for queue in queues
        )

    return _get_num_queued_messages


@pytest.fixture
def get_num_consumed_messages(
    rabbitmq_management_client: AsyncClient,
) -> Callable[[], Awaitable[int]]:
    """Get number of consumed messages in RabbitMQ AMQP."""

    async def _get_num_consumed_messages() -> int:
        queues = (await rabbitmq_management_client.get("queues")).json()
        return sum(queue.get("message_stats", {}).get("ack", 0) for queue in queues)

    return _get_num_consumed_messages



import enum
import json
from typing import Any
from typing import AsyncIterator
from typing import Dict
from typing import Optional
from typing import TypeVar
from typing import cast
from uuid import uuid4

import httpx
from pydantic import BaseModel
from pydantic.json import pydantic_encoder

from .base_model import UNSET
from .exceptions import GraphQLClientGraphQLMultiError
from .exceptions import GraphQLClientHttpError
from .exceptions import GraphQLClientInvalidMessageFormat
from .exceptions import GraphQlClientInvalidResponseError

try:
    from websockets.client import WebSocketClientProtocol
    from websockets.client import connect as ws_connect
    from websockets.typing import Data
    from websockets.typing import Origin
    from websockets.typing import Subprotocol
except ImportError:
    from contextlib import asynccontextmanager

    @asynccontextmanager  # type: ignore
    async def ws_connect(*args, **kwargs):  # pylint: disable=unused-argument
        raise NotImplementedError("Subscriptions require 'websockets' package.")
        yield  # pylint: disable=unreachable

    WebSocketClientProtocol = Any  # type: ignore
    Data = Any  # type: ignore
    Origin = Any  # type: ignore
    Subprotocol = Any  # type: ignore


Self = TypeVar("Self", bound="AsyncBaseClient")

GRAPHQL_TRANSPORT_WS = "graphql-transport-ws"


class GraphQLTransportWSMessageType(str, enum.Enum):
    CONNECTION_INIT = "connection_init"
    CONNECTION_ACK = "connection_ack"
    PING = "ping"
    PONG = "pong"
    SUBSCRIBE = "subscribe"
    NEXT = "next"
    ERROR = "error"
    COMPLETE = "complete"


class AsyncBaseClient:
    def __init__(
        self,
        url: str = "",
        headers: Optional[Dict[str, str]] = None,
        http_client: Optional[httpx.AsyncClient] = None,
        ws_url: str = "",
        ws_headers: Optional[Dict[str, Any]] = None,
        ws_origin: Optional[str] = None,
        ws_connection_init_payload: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.url = url
        self.headers = headers
        self.http_client = (
            http_client if http_client else httpx.AsyncClient(headers=headers)
        )
        self.ws_url = ws_url
        self.ws_headers = ws_headers or {}
        self.ws_origin = Origin(ws_origin) if ws_origin else None
        self.ws_connection_init_payload = ws_connection_init_payload

    async def __aenter__(self: Self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: object,
        exc_val: object,
        exc_tb: object,
    ) -> None:
        await self.http_client.aclose()

    async def execute(
        self, query: str, variables: Optional[Dict[str, Any]] = None
    ) -> httpx.Response:
        payload: Dict[str, Any] = {"query": query}
        if variables:
            payload["variables"] = self._convert_dict_to_json_serializable(variables)
        content = json.dumps(payload, default=pydantic_encoder)
        return await self.http_client.post(
            url=self.url, content=content, headers={"Content-Type": "application/json"}
        )

    def get_data(self, response: httpx.Response) -> Dict[str, Any]:
        if not response.is_success:
            raise GraphQLClientHttpError(
                status_code=response.status_code, response=response
            )

        try:
            response_json = response.json()
        except ValueError as exc:
            raise GraphQlClientInvalidResponseError(response=response) from exc

        if (not isinstance(response_json, dict)) or ("data" not in response_json):
            raise GraphQlClientInvalidResponseError(response=response)

        data = response_json["data"]
        errors = response_json.get("errors")

        if errors:
            raise GraphQLClientGraphQLMultiError.from_errors_dicts(
                errors_dicts=errors, data=data
            )

        return cast(dict[str, Any], data)

    async def execute_ws(
        self, query: str, variables: Optional[Dict[str, Any]] = None
    ) -> AsyncIterator[Dict[str, Any]]:
        operation_id = str(uuid4())
        async with ws_connect(
            self.ws_url,
            subprotocols=[Subprotocol(GRAPHQL_TRANSPORT_WS)],
            origin=self.ws_origin,
            extra_headers=self.ws_headers,
        ) as websocket:
            await self._send_connection_init(websocket)
            await self._send_subscribe(
                websocket,
                operation_id=operation_id,
                query=query,
                variables=variables,
            )

            async for message in websocket:
                data = await self._handle_ws_message(message, websocket)
                if data:
                    yield data

    def _convert_dict_to_json_serializable(
        self, dict_: Dict[str, Any]
    ) -> Dict[str, Any]:
        return {
            key: self._convert_value(value)
            for key, value in dict_.items()
            if value is not UNSET
        }

    def _convert_value(self, value: Any) -> Any:
        if isinstance(value, BaseModel):
            return value.dict(by_alias=True, exclude_unset=True)
        if isinstance(value, list):
            return [self._convert_value(item) for item in value]
        return value

    async def _send_connection_init(self, websocket: WebSocketClientProtocol) -> None:
        payload: Dict[str, Any] = {
            "type": GraphQLTransportWSMessageType.CONNECTION_INIT.value
        }
        if self.ws_connection_init_payload:
            payload["payload"] = self.ws_connection_init_payload
        await websocket.send(json.dumps(payload))

    async def _send_subscribe(
        self,
        websocket: WebSocketClientProtocol,
        operation_id: str,
        query: str,
        variables: Optional[Dict[str, Any]] = None,
    ) -> None:
        payload: Dict[str, Any] = {
            "id": operation_id,
            "type": GraphQLTransportWSMessageType.SUBSCRIBE.value,
            "payload": {"query": query},
        }
        if variables:
            payload["payload"]["variables"] = self._convert_dict_to_json_serializable(
                variables
            )
        await websocket.send(json.dumps(payload))

    async def _handle_ws_message(
        self, message: Data, websocket: WebSocketClientProtocol
    ) -> Optional[Dict[str, Any]]:
        try:
            message_dict = json.loads(message)
        except json.JSONDecodeError as exc:
            raise GraphQLClientInvalidMessageFormat(message=message) from exc

        type_ = message_dict.get("type")
        payload = message_dict.get("payload", {})

        if not type_ or type_ not in {t.value for t in GraphQLTransportWSMessageType}:
            raise GraphQLClientInvalidMessageFormat(message=message)

        if type_ == GraphQLTransportWSMessageType.NEXT:
            if "data" not in payload:
                raise GraphQLClientInvalidMessageFormat(message=message)
            return cast(Dict[str, Any], payload["data"])

        if type_ == GraphQLTransportWSMessageType.COMPLETE:
            await websocket.close()
        elif type_ == GraphQLTransportWSMessageType.PING:
            await websocket.send(
                json.dumps({"type": GraphQLTransportWSMessageType.PONG.value})
            )
        elif type_ == GraphQLTransportWSMessageType.ERROR:
            raise GraphQLClientGraphQLMultiError.from_errors_dicts(
                errors_dicts=payload, data=message_dict
            )

        return None



from typing import Any
from typing import Optional

from .base_model import BaseModel


class FetchEvent(BaseModel):
    event_fetch: Optional["FetchEventEventFetch"]


class FetchEventEventFetch(BaseModel):
    subject: str
    priority: int
    token: Any


FetchEvent.update_forward_refs()
FetchEventEventFetch.update_forward_refs()



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
import warnings
from asyncio import gather
from asyncio.exceptions import TimeoutError
from collections import ChainMap
from enum import Enum
from functools import partial
from itertools import starmap
from operator import itemgetter
from types import TracebackType
from typing import Any
from typing import ChainMap as tChainMap
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Type
from typing import cast
from uuid import UUID

import aiohttp
from more_itertools import chunked
from more_itertools import one
from more_itertools import unzip
from tenacity import retry
from tenacity import stop_after_delay
from tenacity import wait_exponential

from fastramqpi.ra_utils.syncable import Syncable

retry_max_time = 30

# TODO: Pydantic type: #45518
AddressReply = Dict[str, Any]


class AddressType(str, Enum):
    """Enumeration of possible address-types / endpoints."""

    ADDRESS = "adresser"
    """Addresses are the main entry / lookup in DAR.

    A single address may have multiple `ACCESS_ADDRESS`es.
    """

    ACCESS_ADDRESS = "adgangsadresser"
    """Access addresses are sub entries in DAR.

    Multiple access addresses may share a single `Address`, for instance:

    Example:
        `ADDRESS`: `ApartmentComplex 1`, and `ACCESS_ADDRESS`es:

        * `ApartmentComplex 1, tv`
        * `ApartmentComplex 1, th`
        * `ApartmentComplex 1, lf`
    """

    HISTORIC_ADDRESS = "historik/adresser"
    """Historic version of `ADDRESS`"""

    HISTORIC_ACCESS_ADDRESS = "historik/adgangsadresser"
    """Historic version of `ACCESS_ADDRESS`"""


ALL_ADDRESS_TYPES = list(AddressType)


class AsyncDARClient:
    """Asynchronous DAR client.

    Example:
        ```Python
        from os2mo_dar_client import AsyncDARClient

        adarclient = AsyncDARClient()
        async with adarclient:
            print(await adarclient.healthcheck())
        ```
    """

    # TODO: Query endpoints ala dawa_helper.py: #45522
    # TODO: Autocomplete endpoints ala OS2mo: #45521
    # TODO: Caching: #45519

    def __init__(self, timeout: int = 10) -> None:
        """Construct an async DAR client.

        Args:
            timeout: Maximum waiting time for response.
        """
        self._timeout: int = timeout

        self._session: Optional[aiohttp.ClientSession] = None
        self._baseurl: str = "https://api.dataforsyningen.dk"

    async def __aenter__(self) -> "AsyncDARClient":
        await self.aopen()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> bool:
        await self.aclose()
        return False

    async def aopen(self) -> None:
        if self._session:
            warnings.warn("aopen called with existing session", UserWarning)
            return
        connector = aiohttp.TCPConnector(limit=10)
        self._session = aiohttp.ClientSession(connector=connector)

    async def aclose(self) -> None:
        if self._session is None:
            warnings.warn("aclose called without session", UserWarning)
            return
        await self._session.close()
        self._session = None

    def _get_session(self) -> aiohttp.ClientSession:
        if self._session is None:
            raise ValueError("Session not set")
        return self._session

    async def healthcheck(self, timeout: Optional[int] = None) -> bool:
        """Check whether DAR can be reached

        Args:
            timeout: Maximum waiting time for response, defaults to class timeout.

        Returns:
            `True` if reachable, `False` otherwise.
        """
        url = f"{self._baseurl}/autocomplete"
        try:
            async with self._get_session().get(
                url, timeout=aiohttp.ClientTimeout(timeout or self._timeout)
            ) as response:
                if response.status == 200:
                    return True
                return False
        except aiohttp.ClientError:
            return False
        except TimeoutError:
            return False

    async def _cleanse_single(
        self, address_string: str, addrtype: AddressType
    ) -> AddressReply:
        """Request that DAR cleanse a string address for us.

        The cleansing process attempts to convert an unstructured text string into a
        stuctured address reply. The process may fail if the input string is too vague.

        Args:
            address_string: The address string we wish to cleanse.
            addrtype: The address type to lookup.

        Raises:
            aiohttp.ClientResponseError: If anything goes wrong.
            ValueError: If passed a historic addrtype.
            RuntimeError: If one unique match could not be found.

        Returns:
            * dict: DAR Reply
        """
        if addrtype in [
            AddressType.HISTORIC_ADDRESS,
            AddressType.HISTORIC_ACCESS_ADDRESS,
        ]:
            raise ValueError("DAR does not support historic cleansing")

        url = f"{self._baseurl}/datavask/{addrtype.value}"
        params: dict[str, str] = {"betegnelse": address_string}

        async with self._get_session().get(
            url, params=params, timeout=aiohttp.ClientTimeout(self._timeout)
        ) as response:
            response.raise_for_status()
            payload = await response.json()
            # Check match category:
            # A is a near perfect match,
            # B is a unique match,
            # C is a non-unique match (which we do not accept)
            if payload["kategori"] not in ["A", "B"]:
                raise RuntimeError("DAR was unable to find a conclusive match")
            address = one(payload["resultater"])["adresse"]
            return cast(AddressReply, address)

    # TODO: Caching goes in here
    async def _address_fetched(self, uuid: UUID, reply: Dict[str, Any]) -> None:
        pass

    @retry(
        reraise=True,
        wait=wait_exponential(multiplier=1, min=4, max=10),
        stop=stop_after_delay(retry_max_time),
    )
    async def _fetch_single(self, uuid: UUID, addrtype: AddressType) -> AddressReply:
        """Lookup uuid in DAR.

        Args:
            uuid: DAR UUID.
            addrtypes: The address type to lookup.

        Raises:
            aiohttp.ClientResponseError: If anything goes wrong

        Returns:
            * dict: DAR Reply
        """

        url = f"{self._baseurl}/{addrtype.value}/{str(uuid)}"
        params: dict[str, str | int] = {"struktur": "mini", "noformat": 1}

        async with self._get_session().get(
            url, params=params, timeout=aiohttp.ClientTimeout(self._timeout)
        ) as response:
            response.raise_for_status()
            payload = await response.json()
            return cast(AddressReply, payload)

    @retry(
        reraise=True,
        wait=wait_exponential(multiplier=1, min=4, max=10),
        stop=stop_after_delay(retry_max_time),
    )
    async def _fetch_non_chunked(
        self, uuids: Set[UUID], addrtype: AddressType
    ) -> Tuple[Dict[UUID, AddressReply], Set[UUID]]:
        """Lookup uuids in DAR (no chunking).

        Args:
            uuids: List of DAR UUIDs to lookup.
            addrtype: The address type to lookup.

        Returns:
            * dict: Map from UUID to DAR reply.
            * set: Set of UUIDs of entries which were not found.
        """
        url = f"{self._baseurl}/{addrtype.value}"
        params: dict[str, str | int] = {
            "id": "|".join(map(str, uuids)),
            "struktur": "mini",
            "noformat": 1,
        }

        async with self._get_session().get(
            url, params=params, timeout=aiohttp.ClientTimeout(self._timeout)
        ) as response:
            response.raise_for_status()
            body = await response.json()

            result_uuids = map(UUID, map(itemgetter("id"), body))
            result = dict(zip(result_uuids, body))

            found_uuids = result.keys()
            missing = set(uuids) - found_uuids

            return result, missing

    async def _fetch_chunked(
        self, uuids: Set[UUID], addrtype: AddressType, chunk_size: int
    ) -> Tuple[Dict[UUID, AddressReply], Set[UUID]]:
        """Lookup uuids in DAR (chunked).

        Chunks the UUIDs before calling `_fetch_non_chunked` on each chunk.

        Args:
            uuids: List of DAR UUIDs.
            addrtype: The address type to lookup.
            chunk_size: Number of UUIDs per block, sent to DAR.

        Returns:
            * dict: Map from UUID to DAR reply.
            * set: Set of UUIDs of entries which were not found.
        """

        # Chunk our UUIDs into blocks of chunk_size
        uuid_chunks = chunked(uuids, chunk_size)
        # Convert chunks into a list of asyncio.tasks
        tasks = map(partial(self._fetch_non_chunked, addrtype=addrtype), uuid_chunks)  # type: ignore
        # Here 'result' is a list of tuples (dict, set) => (result, missing)
        result = await gather(*tasks)
        # First we unzip 'result' to get a list of results and a list of missing
        result_dicts, missing_sets = unzip(result)
        # Then we union the dicts and sets before returning
        combined_result = dict(ChainMap(*result_dicts))  # type: ignore
        combined_missing = set.union(*missing_sets)  # type: ignore
        return combined_result, combined_missing  # type: ignore

    async def _fetch(
        self, uuids: Set[UUID], addrtype: AddressType, chunk_size: int
    ) -> Tuple[Dict[UUID, AddressReply], Set[UUID]]:
        """Lookup uuids in DAR (chunked if required).

        Args:
            uuids: List of DAR UUIDs.
            addrtype: The address type to lookup.
            chunk_size: Number of UUIDs per block, sent to DAR.

        Returns:
            * dict: Map from UUID to DAR reply.
            * set: Set of UUIDs of entries which were not found.
        """
        num_uuids = len(uuids)
        # Short-circuit if possible, chunk if required
        if num_uuids == 0:
            return dict(), set()
        if num_uuids <= chunk_size:
            res: Tuple[
                Dict[UUID, AddressReply], Set[UUID]
            ] = await self._fetch_non_chunked(uuids, addrtype)
            return res
        return await self._fetch_chunked(uuids, addrtype, chunk_size)

    async def fetch(
        self,
        uuids: Set[UUID],
        addrtypes: Optional[List[AddressType]] = None,
        # WARNING: DAR does not support paths of more than 4096 characters on
        # HTTP/1.1. aiohttp does not support HTTP/2. Do not increase the
        # `chunk_size` without testing irl.
        chunk_size: int = 100,
    ) -> Tuple[Dict[UUID, AddressReply], Set[UUID]]:
        """Lookup uuids in DAR (chunked if necessary).

        Calls `_fetch` with all the provided addrtypes.

        Args:
            uuids: List of DAR UUIDs.
            addrtypes: The address type(s) to lookup. If `None` all 4 types are checked.
            chunk_size: Number of UUIDs per block, sent to DAR.

        Returns:
            * dict: Map from UUID to DAR reply.
            * set: Set of UUIDs of entries which were not found.
        """
        addrtypes = addrtypes or ALL_ADDRESS_TYPES
        combined_result: tChainMap[UUID, AddressReply] = ChainMap({})
        # TODO: Do all 4 in parallel?
        for addrtype in addrtypes:
            result, missing = await self._fetch(uuids, addrtype, chunk_size=chunk_size)
            combined_result = ChainMap(combined_result, result)
            # If we managed to find everything, there is no need to check the remaining
            # address types, we can simply return our findings
            if not missing:
                break
            # We only need to check the missing UUIDs in the remaining address types
            uuids = missing
        final_result = dict(combined_result)
        await gather(*starmap(self._address_fetched, final_result.items()))
        return final_result, missing

    async def fetch_single(
        self, uuid: UUID, addrtypes: Optional[List[AddressType]] = None
    ) -> AddressReply:
        """Lookup uuid in DAR.

        Calls `_fetch_single` with all the provided addrtypes.

        Args:
            uuid: DAR UUID.
            addrtypes: The address type(s) to lookup. If `None` all 4 types are checked.

        Raises:
            ValueError: If no match could be found

        Returns:
            * dict: DAR Reply
        """
        addrtypes = addrtypes or ALL_ADDRESS_TYPES
        # TODO: Do all 4 in parallel?
        for addrtype in addrtypes:
            try:
                payload: AddressReply = await self._fetch_single(uuid, addrtype)
                # If we get here everything went well
                await self._address_fetched(uuid, payload)
                return payload
            except aiohttp.ClientResponseError as exc:
                # If not found, try the next address type
                if exc.status == 404:
                    continue
                raise exc
        raise ValueError("No address match found in DAR")

    async def cleanse_single(
        self, address_string: str, addrtypes: Optional[List[AddressType]] = None
    ) -> AddressReply:
        """Request that DAR cleanse a string address for us.

        The cleansing process attempts to convert an unstructured text string into a
        stuctured address reply. The process may fail if the input string is too vague.

        Calls `_cleanse_single` with appropriate addrtypes.

        Args:
            address_string: The address string we wish to cleanse.
            addrtypes: The address type(s) to lookup. If `None` all 4 types are checked.

        Raises:
            aiohttp.ClientResponseError: If anything goes wrong.
            ValueError: If passed a historic addrtype.
            RuntimeError: If one unique match could not be found.

        Returns:
            * dict: DAR Reply
        """
        addrtypes = addrtypes or [AddressType.ADDRESS, AddressType.ACCESS_ADDRESS]
        # TODO: Do all in parallel?
        for addrtype in addrtypes:
            try:
                payload = await self._cleanse_single(address_string, addrtype)
                return payload
            except aiohttp.ClientResponseError as exc:
                # If not found, try the next address type
                if exc.status == 404:
                    continue
                raise exc
            except RuntimeError:
                # If we did not find a conclusive match, try the next address type
                continue
        raise ValueError("No address match found from cleansing in DAR")


class DARClient(Syncable, AsyncDARClient):
    """Synchronous DAR client.

        Example:
            ```Python
            from os2mo_dar_client import DARClient

            darclient = DARClient()
            with darclient:
                print(darclient.healthcheck())
            ```

    Is implemented atop the `AsyncDARClient` using `ra_utils.Syncable`.
    """

    pass



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
import logging

import structlog
from structlog.types import Processor


def configure_logging(log_level: str, json_logs: bool = True) -> None:
    # Heavily inspired by OS2mo

    # NOTE: We intentionally do not log timestamps, as we rely on the container runtime
    #       to add timestamps to our logs. With docker simply add `-t` to logs.

    shared_processors: list[Processor] = [
        # Merge asyncio bound logging variables
        structlog.contextvars.merge_contextvars,
        # Add the name of the logger to event dict.
        structlog.stdlib.add_logger_name,
        # Add log level to event dict.
        structlog.stdlib.add_log_level,
        # Add extra attributes to the event dictionary
        structlog.stdlib.ExtraAdder(),
        # If the "stack_info" key in the event dict is true, remove it and
        # render the current stack trace in the "stack" key.
        structlog.processors.StackInfoRenderer(),
        # If the "exc_info" key in the event dict is either true or a
        # sys.exc_info() tuple, remove "exc_info" and render the exception
        # with traceback into the "exception" key.
        structlog.processors.format_exc_info,
        # If some value is in bytes, decode it to a Unicode str.
        structlog.processors.UnicodeDecoder(),
    ]

    structlog.configure(
        processors=[
            *shared_processors,
            # Prepare event dict for `ProcessorFormatter`.
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
        ],
        # `wrapper_class` is the bound logger that you get back from
        # get_logger(). This one imitates the API of `logging.Logger`.
        wrapper_class=structlog.stdlib.BoundLogger,
        # `logger_factory` is used to create wrapped loggers that are used for
        # OUTPUT. This one returns a `logging.Logger`. The final value (a JSON
        # string) from the final processor (`JSONRenderer`) will be passed to
        # the method of the same name as that you've called on the bound logger.
        logger_factory=structlog.stdlib.LoggerFactory(),
        # Cached logger breaks logs capture during testing.
        # https://www.structlog.org/en/stable/testing.html
        cache_logger_on_first_use=False,
    )

    log_renderer: structlog.types.Processor = structlog.processors.JSONRenderer()
    if not json_logs:  # pragma: no cover
        log_renderer = structlog.dev.ConsoleRenderer()

    formatter = structlog.stdlib.ProcessorFormatter(
        # These run ONLY on `logging` entries that do NOT originate within
        # structlog.
        foreign_pre_chain=shared_processors,
        # These run on ALL entries after the pre_chain is done.
        processors=[
            # Remove _record & _from_structlog.
            structlog.stdlib.ProcessorFormatter.remove_processors_meta,
            log_renderer,
        ],
    )

    handler = logging.StreamHandler()
    # Use OUR `ProcessorFormatter` to format all `logging` entries.
    handler.setFormatter(formatter)
    root_logger = logging.getLogger()
    root_logger.handlers = [handler]
    root_logger.setLevel(log_level.upper())



# SPDX-FileCopyrightText: Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
from typing import Any
from typing import List
from typing import Optional

import sentry_sdk
import structlog
from pydantic import BaseSettings
from pydantic import HttpUrl
from pydantic import ValidationError


class Settings(BaseSettings):
    dsn: HttpUrl

    release: Optional[str]
    environment: Optional[str]
    server_name: Optional[str]
    traces_sample_rate: Optional[str]
    max_breadcrumbs: Optional[str]
    debug: Optional[bool]
    attach_stacktrace: Optional[bool]
    integrations: Optional[List[str]]
    # TODO: Add more settings as needed

    class Config:
        env_prefix = "SENTRY_"


def sentry_init(*args: Any, **kwargs: Any) -> bool:
    """Initialize sentry
    Configure with inputs as kwargs or environment variables
    prefixed with SENTRY_
    Only sets up Sentry if sentry_dsn is found.
    Returns True if sentry was initialized and False if not.
    """
    log = structlog.get_logger()
    try:
        settings = Settings(*args, **kwargs)
    except ValidationError:
        log.info("Sentry init skipped")
        return False
    log.info(f"Setting up Sentry with {settings=}")
    sentry_sdk.init(**settings.dict())
    return True
